import asyncio
import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any, Final, Self

from aiotools import cancel_and_wait
from glide import (
    AdvancedGlideClientConfiguration,
    GlideClient,
    GlideClientConfiguration,
    Logger,
    LogLevel,
    NodeAddress,
    ServerCredentials,
    TlsAdvancedConfiguration,
)
from glide.exceptions import ClosingError  # type: ignore[import-not-found]
from redis.asyncio.sentinel import Sentinel

from ai.backend.common.exception import ClientNotConnectedError, ValkeySentinelMasterNotFound
from ai.backend.common.types import ValkeyTarget
from ai.backend.common.utils import addr_to_hostport_pair
from ai.backend.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


_DEFAULT_REQUEST_TIMEOUT: Final[int] = 1_000  # Default request timeout in milliseconds
_MONITOR_REQUEST_TIMEOUT: Final[int] = 3_000  # Fixed timeout for monitor client in milliseconds
_DEFAULT_CONSECUTIVE_FAILURE_THRESHOLD: Final[int] = (
    3  # Number of consecutive failures before reconnection
)
_DEFAULT_MONITOR_INTERVAL: Final[float] = 10.0  # Interval between ping attempts in seconds

Logger.init(LogLevel.OFF)  # Disable Glide logging by default


SSL_CERT_NONE = "none"
SSL_CERT_REQUIRED = "required"


@dataclass
class ValkeyStandaloneTarget:
    address: str
    password: str | None = None
    request_timeout: int | None = None
    use_tls: bool = False
    tls_skip_verify: bool = False

    @classmethod
    def from_valkey_target(cls, valkey_target: ValkeyTarget) -> Self:
        """
        Create a ValkeyConnectionConfig from a RedisTarget.
        """
        if not valkey_target.addr:
            raise ValueError("RedisTarget must have an address for standalone mode")
        return cls(
            address=valkey_target.addr,
            password=valkey_target.password,
            request_timeout=valkey_target.request_timeout,
            use_tls=valkey_target.use_tls,
            tls_skip_verify=valkey_target.tls_skip_verify,
        )


@dataclass
class ValkeySentinelTarget:
    sentinel_addresses: Iterable[str]
    service_name: str
    password: str | None = None
    request_timeout: int | None = None
    use_tls: bool = False
    tls_skip_verify: bool = False

    @classmethod
    def from_valkey_target(cls, valkey_target: ValkeyTarget) -> Self:
        """
        Create a ValkeySentinelTarget from a ValkeyTarget.
        """
        if not valkey_target.service_name:
            raise ValueError("RedisTarget must have service_name when using sentinel")

        if not valkey_target.sentinel:
            raise ValueError("RedisTarget sentinel configuration is invalid or empty")

        return cls(
            sentinel_addresses=valkey_target.sentinel,
            service_name=valkey_target.service_name,
            password=valkey_target.password,
            request_timeout=valkey_target.request_timeout,
            use_tls=valkey_target.use_tls,
            tls_skip_verify=valkey_target.tls_skip_verify,
        )


class AbstractValkeyClient(ABC):
    @property
    @abstractmethod
    def client(self) -> GlideClient:
        pass

    @abstractmethod
    async def connect(self) -> None:
        pass

    @abstractmethod
    async def disconnect(self) -> None:
        pass

    @abstractmethod
    async def ping(self) -> None:
        """
        Ping the server to check if the connection is alive.

        Raises:
            Exception: If the ping fails or connection is not available
        """
        pass

    @abstractmethod
    async def need_reconnect(self) -> bool:
        """
        Check if reconnection is needed.

        For Sentinel clients, this checks if the master address has changed.
        For standalone clients, this returns True only if the client is not connected.
        """
        raise NotImplementedError


class ValkeyStandaloneClient(AbstractValkeyClient):
    _target: ValkeyStandaloneTarget
    _valkey_client: GlideClient | None
    _db_id: int
    _pubsub_channels: set[str] | None
    _human_readable_name: str

    def __init__(
        self,
        target: ValkeyStandaloneTarget,
        db_id: int,
        human_readable_name: str,
        pubsub_channels: set[str] | None = None,
    ) -> None:
        self._target = target
        self._valkey_client = None
        self._db_id = db_id
        self._human_readable_name = human_readable_name
        self._pubsub_channels = pubsub_channels

    @property
    def client(self) -> GlideClient:
        if self._valkey_client is None:
            raise ClientNotConnectedError("ValkeyStandaloneClient is not connected")
        return self._valkey_client

    async def connect(self) -> None:
        if self._valkey_client is not None:
            return

        await self._create_valkey_client()

    async def disconnect(self) -> None:
        if self._valkey_client:
            await self._valkey_client.close(err_message="ValkeyStandaloneClient is closed.")
            self._valkey_client = None

    async def _create_valkey_client(self) -> None:
        target_host, target_port = addr_to_hostport_pair(self._target.address)
        addresses = [NodeAddress(host=target_host, port=target_port)]

        credentials = (
            ServerCredentials(password=self._target.password) if self._target.password else None
        )

        config = GlideClientConfiguration(
            addresses,
            credentials=credentials,
            database_id=self._db_id,
            request_timeout=self._target.request_timeout or _DEFAULT_REQUEST_TIMEOUT,
            pubsub_subscriptions=GlideClientConfiguration.PubSubSubscriptions(
                channels_and_patterns={
                    GlideClientConfiguration.PubSubChannelModes.Exact: self._pubsub_channels,
                },
                callback=None,
                context=None,
            )
            if self._pubsub_channels
            else None,
            use_tls=self._target.use_tls,
            advanced_config=AdvancedGlideClientConfiguration(
                tls_config=TlsAdvancedConfiguration(
                    use_insecure_tls=self._target.tls_skip_verify,
                ),
            ),
        )

        glide_client = await GlideClient.create(config)
        self._valkey_client = glide_client

        log.debug(
            "Created ValkeyClient for standalone at {}:{} for database {}",
            target_host,
            target_port,
            self._human_readable_name,
        )

    async def ping(self) -> None:
        """
        Ping the server to check if the connection is alive.

        Raises:
            ClientNotConnectedError: If the client is not connected
            Exception: If the ping fails
        """
        if self._valkey_client is None:
            raise ClientNotConnectedError("ValkeyStandaloneClient is not connected")
        await self._valkey_client.ping()

    async def need_reconnect(self) -> bool:
        return self._valkey_client is None


class ValkeySentinelClient(AbstractValkeyClient):
    _target: ValkeySentinelTarget
    _db_id: int
    _human_readable_name: str
    _sentinel: Sentinel
    _pubsub_channels: set[str] | None
    _valkey_client: GlideClient | None
    _master_address: tuple[str, int] | None

    def __init__(
        self,
        target: ValkeySentinelTarget,
        db_id: int,
        human_readable_name: str,
        pubsub_channels: set[str] | None = None,
    ) -> None:
        sentinel_addrs = []
        for addr in target.sentinel_addresses:
            host, port = addr_to_hostport_pair(addr)
            sentinel_addrs.append((host, port))

        sentinel_kwargs: dict[str, Any] = {
            "password": target.password,
            "ssl": target.use_tls,
            "ssl_cert_reqs": SSL_CERT_NONE if target.tls_skip_verify else SSL_CERT_REQUIRED,
        }
        self._sentinel = Sentinel(
            sentinel_addrs,
            sentinel_kwargs=sentinel_kwargs,
        )
        self._db_id = db_id
        self._human_readable_name = human_readable_name
        self._target = target
        self._pubsub_channels = pubsub_channels
        self._valkey_client = None
        self._master_address = None

    @property
    def client(self) -> GlideClient:
        if self._valkey_client is None:
            raise ClientNotConnectedError("ValkeySentinelClient is not connected")
        return self._valkey_client

    async def connect(self) -> None:
        if self._valkey_client is not None:
            return

        await self._create_valkey_client()

    async def disconnect(self) -> None:
        if self._valkey_client:
            await self._valkey_client.close(err_message="ValkeySentinelClient is closed.")
            self._valkey_client = None

    async def _create_valkey_client(self) -> None:
        master_address = await self._get_master_address()
        if master_address is None:
            raise ValkeySentinelMasterNotFound(
                f"Cannot find master for service '{self._target.service_name}'"
            )

        self._master_address = master_address

        addresses = [NodeAddress(host=master_address[0], port=master_address[1])]
        credentials = (
            ServerCredentials(password=self._target.password) if self._target.password else None
        )

        config = GlideClientConfiguration(
            addresses,
            credentials=credentials,
            database_id=self._db_id,
            client_name=self._target.service_name,
            request_timeout=self._target.request_timeout or _DEFAULT_REQUEST_TIMEOUT,
            pubsub_subscriptions=GlideClientConfiguration.PubSubSubscriptions(
                channels_and_patterns={
                    GlideClientConfiguration.PubSubChannelModes.Exact: self._pubsub_channels,
                },
                callback=None,
                context=None,
            )
            if self._pubsub_channels
            else None,
            use_tls=self._target.use_tls,
            advanced_config=AdvancedGlideClientConfiguration(
                tls_config=TlsAdvancedConfiguration(
                    use_insecure_tls=self._target.tls_skip_verify,
                ),
            ),
        )

        glide_client = await GlideClient.create(config)
        self._valkey_client = glide_client

        log.info(
            "Created ValkeyClient for master at {}:{} for database {}",
            master_address[0],
            master_address[1],
            self._human_readable_name,
        )

    async def _get_master_address(self) -> tuple[str, int] | None:
        try:
            return await self._sentinel.discover_master(self._target.service_name)
        except Exception as e:
            log.error(
                "Failed to discover master for service '{}': {}", self._target.service_name, e
            )
            return None

    async def ping(self) -> None:
        """
        Ping the current master to check if the connection is alive.

        Raises:
            ClientNotConnectedError: If the client is not connected
            Exception: If the ping fails
        """
        if self._valkey_client is None:
            raise ClientNotConnectedError("ValkeySentinelClient is not connected")
        await self._valkey_client.ping()

    async def need_reconnect(self) -> bool:
        """Check if reconnection is needed (client disconnected or master changed)."""
        if self._valkey_client is None:
            return True

        if self._master_address is None:
            return True

        current_master = await self._get_master_address()
        if current_master is None:
            return False

        return current_master != self._master_address


def _create_valkey_client_internal(
    valkey_target: ValkeyTarget,
    db_id: int,
    human_readable_name: str,
    pubsub_channels: set[str] | None = None,
) -> AbstractValkeyClient:
    """
    Internal helper to create a basic Valkey client (Standalone or Sentinel).
    """
    if valkey_target.sentinel:
        sentinel_target = ValkeySentinelTarget.from_valkey_target(valkey_target)
        return ValkeySentinelClient(sentinel_target, db_id, human_readable_name, pubsub_channels)
    standalone_target = ValkeyStandaloneTarget.from_valkey_target(valkey_target)
    return ValkeyStandaloneClient(standalone_target, db_id, human_readable_name, pubsub_channels)


def create_valkey_client(
    valkey_target: ValkeyTarget,
    db_id: int,
    human_readable_name: str,
    pubsub_channels: set[str] | None = None,
) -> AbstractValkeyClient:
    """
    Factory function to create a Valkey client based on the target type.

    Returns MonitoringValkeyClient that wraps separate work and monitor clients
    for improved reliability with long-running operations.
    """
    # Create operation client with user-specified timeout
    operation_client = _create_valkey_client_internal(
        valkey_target, db_id, human_readable_name, pubsub_channels
    )

    # Create monitor client with fixed 3-second timeout
    monitor_target = ValkeyTarget(
        addr=valkey_target.addr,
        sentinel=valkey_target.sentinel,
        service_name=valkey_target.service_name,
        password=valkey_target.password,
        request_timeout=_MONITOR_REQUEST_TIMEOUT,
        use_tls=valkey_target.use_tls,
        tls_skip_verify=valkey_target.tls_skip_verify,
    )
    monitor_client = _create_valkey_client_internal(
        monitor_target, db_id, f"{human_readable_name}-monitor", None
    )

    return MonitoringValkeyClient(operation_client, monitor_client)


class MonitoringValkeyClient(AbstractValkeyClient):
    """
    Valkey client wrapper with separated monitor client for health checks.

    This client wraps two separate Valkey clients:
    - operation_client: For actual operations (user-specified timeout)
    - monitor_client: For health checks only (fixed 3-second timeout)

    This separation prevents timeout issues when performing long-running operations
    like stream reads while maintaining connection health monitoring.
    """

    _operation_client: AbstractValkeyClient
    _monitor_client: AbstractValkeyClient
    _monitor_task: asyncio.Task[None] | None
    _reconnectable_exceptions: tuple[type[Exception], ...]
    _consecutive_failure_threshold: int
    _consecutive_failure_count: int

    def __init__(
        self,
        operation_client: AbstractValkeyClient,
        monitor_client: AbstractValkeyClient,
        reconnectable_exceptions: tuple[type[Exception], ...] = (
            ClosingError,
            ClientNotConnectedError,
        ),
        consecutive_failure_threshold: int = _DEFAULT_CONSECUTIVE_FAILURE_THRESHOLD,
    ) -> None:
        self._operation_client = operation_client
        self._monitor_client = monitor_client
        self._monitor_task = None
        self._reconnectable_exceptions = reconnectable_exceptions
        self._consecutive_failure_threshold = consecutive_failure_threshold
        self._consecutive_failure_count = 0
        self._closed = False

    @property
    def client(self) -> GlideClient:
        return self._operation_client.client

    async def connect(self) -> None:
        await self._operation_client.connect()
        await self._monitor_client.connect()
        self._monitor_task = asyncio.create_task(self._monitor_connection())

    async def disconnect(self) -> None:
        self._closed = True
        if self._monitor_task:
            await cancel_and_wait(self._monitor_task)
            self._monitor_task = None

        await self._monitor_client.disconnect()
        await self._operation_client.disconnect()

    async def ping(self) -> None:
        """
        Ping the server to check if the connection is alive.
        Uses the monitor client to avoid interfering with operation tasks.

        Raises:
            ClientNotConnectedError: If the client is not connected
            Exception: If the ping fails
        """
        await self._monitor_client.ping()

    async def need_reconnect(self) -> bool:
        return await self._monitor_client.need_reconnect()

    async def _check_ping(self) -> bool:
        """
        Ping the monitor client and determine if reconnection is needed.

        Returns:
            True if reconnection is needed, False otherwise.
        """
        reconnectable_exceptions = self._reconnectable_exceptions
        try:
            await self._monitor_client.ping()
            self._consecutive_failure_count = 0
            return False
        except reconnectable_exceptions as e:
            log.warning("Connection error: {}, reconnecting immediately...", e)
            self._consecutive_failure_count = 0
            return True
        except Exception as e:
            self._consecutive_failure_count += 1
            log.warning(
                "Error in connection monitoring (consecutive failures: {}/{}): {}",
                self._consecutive_failure_count,
                self._consecutive_failure_threshold,
                e,
            )
            if self._consecutive_failure_count >= self._consecutive_failure_threshold:
                log.warning(
                    "Consecutive failure threshold reached ({}), reconnecting...",
                    self._consecutive_failure_threshold,
                )
                self._consecutive_failure_count = 0
                return True
            return False

    async def _check_connection(self) -> bool:
        """
        Check if reconnection is needed by ping and need_reconnect.

        Returns:
            True if reconnection is needed, False otherwise.
        """
        if await self._check_ping():
            return True

        if await self._monitor_client.need_reconnect():
            log.info("Reconnection needed (master changed), reconnecting...")
            return True

        return False

    async def _monitor_connection(self) -> None:
        log.info("Starting Valkey connection monitor task...")
        try:
            while True:
                try:
                    await asyncio.sleep(_DEFAULT_MONITOR_INTERVAL)
                    if await self._check_connection():
                        log.info("Reconnecting Valkey clients...")
                        await self._reconnect()
                except asyncio.CancelledError:
                    # Normal shutdown - don't log as error
                    raise
                except Exception as e:
                    if not self._closed:
                        log.exception("Error in Valkey connection monitor: {}", e)
                        continue
                    raise
        finally:
            log.info("Valkey connection monitor task stopped. Client closed: {}", self._closed)

    async def _reconnect(self) -> None:
        # Disconnect both clients
        try:
            await self._monitor_client.disconnect()
        except Exception as e:
            log.warning("Error disconnecting monitor client: {}", e)

        try:
            await self._operation_client.disconnect()
        except Exception as e:
            log.warning("Error disconnecting operation client: {}", e)

        # Reconnect both clients
        await self._operation_client.connect()
        await self._monitor_client.connect()



"""
Response DTOs for auth system.
Shared between Client SDK and Manager API.
"""

from __future__ import annotations

from pydantic import Field

from ai.backend.common.api_handlers import BaseResponseModel

from .types import AuthSuccessResponse

__all__ = (
    "AuthorizeResponse",
    "GetRoleResponse",
    "SignupResponse",
    "SignoutResponse",
    "UpdateFullNameResponse",
    "UpdatePasswordResponse",
    "UpdatePasswordNoAuthResponse",
    "GetSSHKeypairResponse",
    "SSHKeypairResponse",
)


class AuthorizeResponse(BaseResponseModel):
    """Response for successful authorization."""

    data: AuthSuccessResponse = Field(
        description="Issued credentials (access_key/secret_key), user role, and account status",
    )


class GetRoleResponse(BaseResponseModel):
    """Response containing the user's roles at different scopes."""

    global_role: str = Field(description="System-wide role: 'superadmin' or 'user'")
    domain_role: str = Field(description="Domain-level role: 'admin' or 'user'")
    group_role: str | None = Field(
        default=None,
        description="Group-level role if group ID was requested; None if not a member",
    )


class SignupResponse(BaseResponseModel):
    """Response for a successful signup with auto-generated API keypair."""

    access_key: str = Field(description="Auto-generated API access key")
    secret_key: str = Field(description="Auto-generated API secret key paired with access_key")


class SignoutResponse(BaseResponseModel):
    """Response for signout. Empty body; user and keypairs are now INACTIVE."""

    pass


class UpdateFullNameResponse(BaseResponseModel):
    """Response for update full name (empty response)."""

    pass


class UpdatePasswordResponse(BaseResponseModel):
    """Response for update password."""

    error_msg: str | None = Field(
        default=None,
        description="Error detail on failure (e.g., new password mismatch); None on success",
    )


class UpdatePasswordNoAuthResponse(BaseResponseModel):
    """Response for updating an expired password without API auth."""

    password_changed_at: str = Field(
        description="ISO 8601 timestamp of when the password was changed",
    )


class GetSSHKeypairResponse(BaseResponseModel):
    """Response for retrieving the SSH keypair (public key only)."""

    ssh_public_key: str = Field(description="SSH public key associated with the access key")


class SSHKeypairResponse(BaseResponseModel):
    """Response for generate/upload SSH keypair (both keys returned)."""

    ssh_public_key: str = Field(description="SSH public key in standard format")
    ssh_private_key: str = Field(description="SSH private key in PEM format")



"""
Response DTOs for the CSV export system.

This module defines the response structures returned by the export API endpoints,
including lists of available reports and individual report details.
"""

from __future__ import annotations

from pydantic import Field

from ai.backend.common.api_handlers import BaseResponseModel

from .types import ExportReportInfo

__all__ = (
    "GetExportReportResponse",
    "ListExportReportsResponse",
)


class ListExportReportsResponse(BaseResponseModel):
    """
    Response containing a list of all available export reports.

    Returned by GET /export/reports endpoint.
    Use this to discover what reports are available for export.
    """

    reports: list[ExportReportInfo] = Field(
        description=(
            "List of all available export reports. "
            "Each report includes its key (for use in export requests), "
            "name, description, and list of exportable fields."
        )
    )


class GetExportReportResponse(BaseResponseModel):
    """
    Response containing details of a single export report.

    Returned by GET /export/reports/{report_key} endpoint.
    Use this to get detailed information about a specific report
    before making an export request.
    """

    report: ExportReportInfo = Field(
        description=(
            "Detailed information about the requested report, "
            "including its metadata and all available fields for export."
        )
    )



"""
Common DTOs for deployment system used by both Client SDK and Manager.
"""

from __future__ import annotations

from .request import (
    BlueGreenConfigInput,
    ClusterConfigInput,
    CreateDeploymentRequest,
    CreateRevisionRequest,
    DeploymentFilter,
    DeploymentMetadataInput,
    DeploymentPathParam,
    DeploymentStrategyInput,
    ExtraVFolderMountInput,
    ImageInput,
    ModelMountConfigInput,
    ModelRuntimeConfigInput,
    NetworkAccessInput,
    ResourceConfigInput,
    RevisionFilter,
    RevisionInput,
    RevisionPathParam,
    RollingUpdateConfigInput,
    RouteFilter,
    RoutePathParam,
    SearchDeploymentsRequest,
    SearchRevisionsRequest,
    SearchRoutesRequest,
    UpdateDeploymentRequest,
    UpdateRouteTrafficStatusRequest,
)
from .response import (
    ActivateRevisionResponse,
    ClusterConfigDTO,
    CreateDeploymentResponse,
    CreateRevisionResponse,
    CursorPaginationInfo,
    DeactivateRevisionResponse,
    DeploymentDTO,
    DestroyDeploymentResponse,
    GetDeploymentResponse,
    GetRevisionResponse,
    ListDeploymentsResponse,
    ListRevisionsResponse,
    ListRoutesResponse,
    ModelMountConfigDTO,
    ModelRuntimeConfigDTO,
    NetworkConfigDTO,
    PaginationInfo,
    ReplicaStateDTO,
    ResourceConfigDTO,
    RevisionDTO,
    RouteDTO,
    UpdateDeploymentResponse,
    UpdateRouteTrafficStatusResponse,
)
from .types import (
    DeploymentOrder,
    DeploymentOrderField,
    OrderDirection,
    RevisionOrder,
    RevisionOrderField,
    RouteOrder,
    RouteOrderField,
)

__all__ = (
    # Types
    "OrderDirection",
    "DeploymentOrderField",
    "RevisionOrderField",
    "RouteOrderField",
    "DeploymentOrder",
    "RevisionOrder",
    "RouteOrder",
    # Request DTOs - Filters
    "DeploymentFilter",
    "RevisionFilter",
    "RouteFilter",
    # Request DTOs - Path params
    "DeploymentPathParam",
    "RevisionPathParam",
    "RoutePathParam",
    # Request DTOs - Search/List
    "SearchDeploymentsRequest",
    "SearchRevisionsRequest",
    "SearchRoutesRequest",
    # Request DTOs - Create inputs
    "RollingUpdateConfigInput",
    "BlueGreenConfigInput",
    "DeploymentMetadataInput",
    "NetworkAccessInput",
    "DeploymentStrategyInput",
    "ImageInput",
    "ClusterConfigInput",
    "ResourceConfigInput",
    "ModelMountConfigInput",
    "ModelRuntimeConfigInput",
    "ExtraVFolderMountInput",
    "RevisionInput",
    # Request DTOs - Create/Update requests
    "CreateDeploymentRequest",
    "CreateRevisionRequest",
    "UpdateDeploymentRequest",
    "UpdateRouteTrafficStatusRequest",
    # Response DTOs - Data
    "DeploymentDTO",
    "RevisionDTO",
    "RouteDTO",
    "NetworkConfigDTO",
    "ClusterConfigDTO",
    "ResourceConfigDTO",
    "ModelRuntimeConfigDTO",
    "ModelMountConfigDTO",
    "ReplicaStateDTO",
    # Response DTOs - Create responses
    "CreateDeploymentResponse",
    "CreateRevisionResponse",
    # Response DTOs - Get/List responses
    "GetDeploymentResponse",
    "ListDeploymentsResponse",
    "UpdateDeploymentResponse",
    "DestroyDeploymentResponse",
    "GetRevisionResponse",
    "ListRevisionsResponse",
    "ActivateRevisionResponse",
    "DeactivateRevisionResponse",
    "ListRoutesResponse",
    "UpdateRouteTrafficStatusResponse",
    # Response DTOs - Pagination
    "PaginationInfo",
    "CursorPaginationInfo",
)



from __future__ import annotations

import uuid
from collections.abc import Mapping
from typing import Any, Self

from pydantic import BaseModel, Field

from ai.backend.common.data.artifact.types import ArtifactRegistryType


class ArtifactRegistryStatefulData(BaseModel):
    """Base class for artifact registry stateful data."""

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Self:
        """Create instance from dictionary with Pydantic validation."""
        return cls.model_validate(data)

    def to_dict(self) -> Mapping[str, Any]:
        """Convert instance to dictionary with JSON-compatible types."""
        return self.model_dump(mode="json")


class _CommonArtifactRegistryStatefulData(ArtifactRegistryStatefulData):
    """Common fields for all artifact registry stateful data types."""

    id: uuid.UUID = Field(
        description="Primary key from the artifact_registries table representing the unified registry entry",
    )
    registry_id: uuid.UUID = Field(
        description=(
            "Foreign key referencing the type-specific registry table "
            "(huggingface_registries or reservoir_registries) containing detailed configuration"
        ),
    )
    name: str = Field(
        description="Human-readable name of the registry used for display and identification purposes",
    )
    type: ArtifactRegistryType = Field(
        description=(
            "Type of the artifact registry (HUGGINGFACE or RESERVOIR) "
            "determining the registry implementation"
        ),
    )


class HuggingFaceRegistryStatefulData(_CommonArtifactRegistryStatefulData):
    """Stateful data type for HuggingFace registry used in Valkey client."""

    url: str = Field(
        description=(
            "Base URL of the HuggingFace Hub instance (e.g., https://huggingface.co) "
            "used for model and dataset downloads"
        ),
    )
    token: str | None = Field(
        default=None,
        description="Optional authentication token for accessing private HuggingFace repositories and datasets",
    )


class ReservoirRegistryStatefulData(_CommonArtifactRegistryStatefulData):
    """Stateful data type for Reservoir registry used in Valkey client."""

    endpoint: str = Field(
        description=(
            "Endpoint URL of the Backend.AI manager instance "
            "configured as a Reservoir registry for artifact storage"
        ),
    )
    access_key: str = Field(
        description=(
            "Access key credential used for authenticating API requests "
            "when communicating with the Backend.AI manager instance configured as Reservoir registry"
        ),
    )
    secret_key: str = Field(
        description=(
            "Secret key credential used for signing and authenticating API requests "
            "when communicating with the Backend.AI manager instance configured as Reservoir registry"
        ),
    )
    api_version: str = Field(
        description="API version string of the Reservoir service protocol (e.g., 'v1') for compatibility management",
    )



from __future__ import annotations

from typing import Annotated

from pydantic import Field

from ai.backend.common.config import BaseConfigSchema
from ai.backend.common.data.config.types import EtcdConfigData
from ai.backend.common.meta import BackendAIConfigMeta, ConfigExample
from ai.backend.common.typed_validators import HostPortPair

__all__ = ("EtcdConfig",)


class EtcdConfig(BaseConfigSchema):
    """Configuration for etcd connection.

    etcd is used as the distributed key-value store for Backend.AI cluster
    configuration and coordination.
    """

    namespace: Annotated[
        str,
        Field(default="local"),
        BackendAIConfigMeta(
            description=(
                "Namespace prefix for etcd keys used by Backend.AI. "
                "Allows multiple Backend.AI clusters to share the same etcd cluster. "
                "All Backend.AI related keys will be stored under this namespace."
            ),
            added_version="22.03.0",
            example=ConfigExample(local="local", prod="backend"),
        ),
    ]
    addr: Annotated[
        HostPortPair | list[HostPortPair],
        Field(default_factory=lambda: HostPortPair(host="127.0.0.1", port=2379)),
        BackendAIConfigMeta(
            description=(
                "Network address of the etcd server. "
                "Default is the standard etcd port on localhost. "
                "In production, should point to one or more etcd instance endpoint(s)."
            ),
            added_version="22.03.0",
            example=ConfigExample(local="127.0.0.1:2379", prod="etcd-cluster:2379"),
        ),
    ]
    user: Annotated[
        str | None,
        Field(default=None),
        BackendAIConfigMeta(
            description=(
                "Username for authenticating with etcd. "
                "Optional if etcd doesn't require authentication. "
                "Should be set along with password for secure deployments."
            ),
            added_version="22.03.0",
            example=ConfigExample(local="", prod="backend"),
        ),
    ]
    password: Annotated[
        str | None,
        Field(default=None),
        BackendAIConfigMeta(
            description=(
                "Password for authenticating with etcd. "
                "Should be kept secret in production environments. "
                "Set together with the user field for authentication."
            ),
            added_version="22.03.0",
            secret=True,
            example=ConfigExample(local="", prod="ETCD_PASSWORD"),
        ),
    ]

    def to_dataclass(self) -> EtcdConfigData:
        """Convert to EtcdConfigData dataclass."""
        return EtcdConfigData(
            namespace=self.namespace,
            addrs=self.addr if isinstance(self.addr, list) else [self.addr],
            user=self.user,
            password=self.password,
        )



"""
Path parameter DTOs for RBAC API endpoints.
Shared between Client SDK and Manager API.
"""

from __future__ import annotations

from uuid import UUID

from pydantic import Field

from ai.backend.common.api_handlers import BaseRequestModel
from ai.backend.common.data.permission.types import EntityType, ScopeType

__all__ = (
    "DeleteObjectPermissionPathParam",
    "DeletePermissionPathParam",
    "GetRolePathParam",
    "SearchEntitiesPathParam",
    "SearchScopesPathParam",
    "SearchUsersAssignedToRolePathParam",
    "UpdateRolePathParam",
)


class GetRolePathParam(BaseRequestModel):
    """Path parameter for getting a role."""

    role_id: UUID = Field(description="The role ID to retrieve")


class UpdateRolePathParam(BaseRequestModel):
    """Path parameter for updating a role."""

    role_id: UUID = Field(description="The role ID to update")


class SearchUsersAssignedToRolePathParam(BaseRequestModel):
    """Path parameter for searching users assigned to a role."""

    role_id: UUID = Field(description="The role ID to search assigned users for")


class DeletePermissionPathParam(BaseRequestModel):
    """Path parameter for deleting a permission."""

    permission_id: UUID = Field(description="The permission ID to delete")


class DeleteObjectPermissionPathParam(BaseRequestModel):
    """Path parameter for deleting an object permission."""

    object_permission_id: UUID = Field(description="The object permission ID to delete")


class SearchScopesPathParam(BaseRequestModel):
    """Path parameter for searching scopes."""

    scope_type: ScopeType = Field(
        description="Scope types", examples=["domain", "project", "user", "global"]
    )


class SearchEntitiesPathParam(BaseRequestModel):
    """Path parameter for searching entities within a scope."""

    scope_type: ScopeType = Field(
        description="Scope type", examples=["domain", "project", "user", "global"]
    )
    scope_id: str = Field(
        description="Scope ID (domain name, project UUID, or user UUID)",
        examples=["default", "550e8400-e29b-41d4-a716-446655440000"],
    )
    entity_type: EntityType = Field(
        description="Entity type to search",
        examples=["user", "vfolder", "session", "image"],
    )



from __future__ import annotations

import hashlib
import logging
import time
from typing import Final, Self, cast

from glide import Batch, ExpirySet, ExpiryType
from pydantic import ValidationError

from ai.backend.common.clients.valkey_client.client import (
    AbstractValkeyClient,
    create_valkey_client,
)
from ai.backend.common.data.artifact.types import (
    ArtifactDownloadTrackingData,
    DownloadProgressData,
    FileDownloadProgressData,
)
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience import (
    BackoffStrategy,
    MetricArgs,
    MetricPolicy,
    Resilience,
    RetryArgs,
    RetryPolicy,
)
from ai.backend.common.types import ValkeyTarget
from ai.backend.logging.utils import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

# Resilience instance for valkey_artifact layer
valkey_artifact_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.VALKEY, layer=LayerType.VALKEY_ARTIFACT)),
        RetryPolicy(
            RetryArgs(
                max_retries=3,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)

_ARTIFACT_DOWNLOAD_PREFIX: Final[str] = "artifact:download"
_ARTIFACT_DOWNLOAD_EXPIRATION: Final[int] = 86400  # 24 hours


class ValkeyArtifactDownloadTrackingClient:
    """
    Client for managing artifact download tracking using Valkey.
    """

    _client: AbstractValkeyClient
    _closed: bool

    def __init__(self, client: AbstractValkeyClient) -> None:
        self._client = client
        self._closed = False

    @classmethod
    async def create(
        cls,
        valkey_target: ValkeyTarget,
        *,
        db_id: int,
        human_readable_name: str,
    ) -> Self:
        """
        Create a ValkeyArtifactDownloadTrackingClient instance.

        :param valkey_target: The target Valkey server to connect to.
        :param db_id: The database index to use.
        :param human_readable_name: The human-readable name of the client.
        :return: An instance of ValkeyArtifactDownloadTrackingClient.
        """
        client = create_valkey_client(
            valkey_target=valkey_target,
            db_id=db_id,
            human_readable_name=human_readable_name,
        )
        await client.connect()
        return cls(client=client)

    @valkey_artifact_resilience.apply()
    async def close(self) -> None:
        """
        Close the ValkeyArtifactDownloadTrackingClient connection.
        """
        if self._closed:
            log.debug("ValkeyArtifactDownloadTrackingClient is already closed.")
            return
        self._closed = True
        await self._client.disconnect()

    async def ping(self) -> None:
        """Ping the Valkey server to check connection health."""
        await self._client.ping()

    def _create_batch(self, is_atomic: bool = False) -> Batch:
        """
        Create a batch for transaction operations.

        :param is_atomic: Whether the batch should be atomic.
        :return: A Batch instance.
        """
        return Batch(is_atomic=is_atomic)

    def _get_artifact_key(self, model_id: str, revision: str) -> str:
        """
        Generate Redis key for artifact-level tracking.

        :param model_id: Model identifier
        :param revision: Model revision
        :return: Redis key string
        """
        # URL-encode model_id and revision to handle special characters
        safe_model_id = model_id.replace("/", ":")
        safe_revision = revision.replace("/", ":")
        return f"{_ARTIFACT_DOWNLOAD_PREFIX}:{safe_model_id}:{safe_revision}"

    def _get_file_key(self, model_id: str, revision: str, file_path: str) -> str:
        """
        Generate Redis key for file-level tracking.

        :param model_id: Model identifier
        :param revision: Model revision
        :param file_path: File path within the model
        :return: Redis key string
        """
        # Hash the file path to handle special characters and length
        file_hash = hashlib.sha256(file_path.encode()).hexdigest()[:16]
        artifact_key = self._get_artifact_key(model_id, revision)
        return f"{artifact_key}:file:{file_hash}"

    def _get_file_pattern(self, model_id: str, revision: str) -> str:
        """
        Generate Redis key pattern for all files in an artifact.

        :param model_id: Model identifier
        :param revision: Model revision
        :return: Redis key pattern string
        """
        artifact_key = self._get_artifact_key(model_id, revision)
        return f"{artifact_key}:file:*"

    @valkey_artifact_resilience.apply()
    async def init_artifact_download(
        self,
        model_id: str,
        revision: str,
        file_info_list: list[tuple[str, int]],
    ) -> None:
        """
        Initialize artifact download tracking in Redis.
        Pre-registers all files with initial progress (0 bytes downloaded).

        :param model_id: Model identifier
        :param revision: Model revision
        :param file_info_list: List of (file_path, file_size) tuples
        """
        artifact_key = self._get_artifact_key(model_id, revision)
        current_time = time.time()

        total_files = len(file_info_list)
        total_bytes = sum(size for _, size in file_info_list)

        batch = self._create_batch(is_atomic=False)

        # Set artifact-level data as Hash for atomic field updates
        batch.hset(
            artifact_key,
            {
                "model_id": model_id,
                "revision": revision,
                "start_time": str(current_time),
                "last_updated": str(current_time),
                "total_files": str(total_files),
                "total_bytes": str(total_bytes),
                "completed_files": "0",
                "downloaded_bytes": "0",
            },
        )
        # Set TTL for artifact hash
        batch.expire(artifact_key, _ARTIFACT_DOWNLOAD_EXPIRATION)

        # Pre-register all files with initial progress (0 bytes)
        for file_path, file_size in file_info_list:
            file_key = self._get_file_key(model_id, revision, file_path)
            file_data = FileDownloadProgressData(
                file_path=file_path,
                success=False,
                current_bytes=0,
                total_bytes=file_size,
                last_updated=current_time,
                error_message=None,
            )
            batch.set(
                file_key,
                file_data.model_dump_json(),
                expiry=ExpirySet(
                    expiry_type=ExpiryType.SEC,
                    value=_ARTIFACT_DOWNLOAD_EXPIRATION,
                ),
            )

        await self._client.client.exec(batch, raise_on_error=True)
        log.debug(
            "Initialized artifact download tracking: model_id={}, revision={}, total_files={}, total_bytes={}",
            model_id,
            revision,
            total_files,
            total_bytes,
        )

    @valkey_artifact_resilience.apply()
    async def update_file_progress(
        self,
        model_id: str,
        revision: str,
        file_path: str,
        current_bytes: int,
        total_bytes: int,
        success: bool = False,
        error_message: str | None = None,
    ) -> None:
        """
        Update file download progress in Redis.

        :param model_id: Model identifier
        :param revision: Model revision
        :param file_path: File path within the model
        :param current_bytes: Current bytes downloaded for this file
        :param total_bytes: Total bytes for this file
        :param success: Whether the download completed successfully
        :param error_message: Error message if download failed
        """
        file_key = self._get_file_key(model_id, revision, file_path)
        artifact_key = self._get_artifact_key(model_id, revision)

        # Get previous file data to calculate delta for artifact aggregation
        previous_data_bytes = await self._client.client.get(file_key)
        previous_current_bytes = 0
        if previous_data_bytes:
            try:
                previous_data = FileDownloadProgressData.model_validate_json(previous_data_bytes)
                previous_current_bytes = previous_data.current_bytes
            except (ValidationError, UnicodeDecodeError):
                log.warning("Failed to parse previous file data for progress update")

        # Calculate bytes delta for artifact aggregation
        bytes_delta = current_bytes - previous_current_bytes

        file_data = FileDownloadProgressData(
            file_path=file_path,
            success=success,
            current_bytes=current_bytes,
            total_bytes=total_bytes,
            last_updated=time.time(),
            error_message=error_message,
        )

        # Update file progress (file key should already exist from init_artifact_download)
        batch = self._create_batch(is_atomic=False)
        batch.set(
            file_key,
            file_data.model_dump_json(),
            expiry=ExpirySet(expiry_type=ExpiryType.KEEP_TTL, value=None),
        )

        await self._client.client.exec(batch, raise_on_error=True)

        # Update artifact aggregates atomically using HINCRBY
        if bytes_delta != 0 or (success and previous_current_bytes < total_bytes):
            current_time_str = str(time.time())

            # Use HINCRBY for atomic increments
            if bytes_delta != 0:
                await self._client.client.hincrby(artifact_key, "downloaded_bytes", bytes_delta)

            if success and previous_current_bytes < total_bytes:
                await self._client.client.hincrby(artifact_key, "completed_files", 1)

            # Update last_updated timestamp
            await self._client.client.hset(artifact_key, {"last_updated": current_time_str})

        log.trace(
            "Updated file progress: file_path={}, current={}, total={}, success={}",
            file_path,
            current_bytes,
            total_bytes,
            success,
        )

    @valkey_artifact_resilience.apply()
    async def get_artifact_progress(
        self,
        model_id: str,
        revision: str,
    ) -> ArtifactDownloadTrackingData | None:
        """
        Get overall artifact download progress.

        :param model_id: Model identifier
        :param revision: Model revision
        :return: Artifact progress data or None if not found
        """
        artifact_key = self._get_artifact_key(model_id, revision)
        hash_data = await self._client.client.hgetall(artifact_key)

        if not hash_data:
            return None

        try:
            # Convert Hash data (bytes keys/values) to proper types
            return ArtifactDownloadTrackingData(
                model_id=hash_data[b"model_id"].decode(),
                revision=hash_data[b"revision"].decode(),
                start_time=float(hash_data[b"start_time"].decode()),
                last_updated=float(hash_data[b"last_updated"].decode()),
                total_files=int(hash_data[b"total_files"].decode()),
                total_bytes=int(hash_data[b"total_bytes"].decode()),
                completed_files=int(hash_data[b"completed_files"].decode()),
                downloaded_bytes=int(hash_data[b"downloaded_bytes"].decode()),
            )
        except (KeyError, ValueError, UnicodeDecodeError) as e:
            log.warning("Failed to parse artifact progress data from hash: {}", str(e))
            return None

    @valkey_artifact_resilience.apply()
    async def get_file_progress(
        self,
        model_id: str,
        revision: str,
        file_path: str,
    ) -> FileDownloadProgressData | None:
        """
        Get specific file download progress.

        :param model_id: Model identifier
        :param revision: Model revision
        :param file_path: File path within the model
        :return: File progress data or None if not found
        """
        file_key = self._get_file_key(model_id, revision, file_path)
        data_bytes = await self._client.client.get(file_key)

        if not data_bytes:
            return None

        try:
            return FileDownloadProgressData.model_validate_json(data_bytes)
        except (ValidationError, UnicodeDecodeError):
            log.warning("Failed to parse file progress data")
            return None

    @valkey_artifact_resilience.apply()
    async def get_all_file_progress(
        self,
        model_id: str,
        revision: str,
    ) -> list[FileDownloadProgressData]:
        """
        Get all file download progress for an artifact.

        :param model_id: Model identifier
        :param revision: Model revision
        :return: List of file progress data
        """
        file_pattern = self._get_file_pattern(model_id, revision)
        file_progress_list: list[FileDownloadProgressData] = []

        cursor = b"0"
        while cursor:
            result = await self._client.client.scan(cursor, match=file_pattern, count=100)
            cursor = cast(bytes, result[0])
            keys = cast(list[bytes], result[1])

            if keys:
                # Get all file progress data
                values = await self._client.client.mget(cast(list[str | bytes], keys))
                for key_bytes, value_bytes in zip(keys, values, strict=False):
                    if value_bytes:
                        try:
                            file_progress = FileDownloadProgressData.model_validate_json(
                                value_bytes
                            )
                            file_progress_list.append(file_progress)
                        except (ValidationError, UnicodeDecodeError):
                            log.warning("Failed to parse file progress data for key: {}", key_bytes)

            if cursor == b"0":
                break

        return file_progress_list

    @valkey_artifact_resilience.apply()
    async def get_download_progress(
        self,
        model_id: str,
        revision: str,
    ) -> DownloadProgressData:
        """
        Get download progress including artifact-level and all file-level data.

        This is a convenience method that combines get_artifact_progress() and
        get_all_file_progress() into a single call.

        :param model_id: Model identifier
        :param revision: Model revision
        :return: Download progress data
        """
        artifact_progress = await self.get_artifact_progress(model_id, revision)
        file_progress = await self.get_all_file_progress(model_id, revision)
        return DownloadProgressData(
            artifact_progress=artifact_progress,
            file_progress=file_progress,
        )

    @valkey_artifact_resilience.apply()
    async def cleanup_artifact_download(
        self,
        model_id: str,
        revision: str,
    ) -> None:
        """
        Clean up all Redis keys for an artifact download.

        :param model_id: Model identifier
        :param revision: Model revision
        """
        artifact_key = self._get_artifact_key(model_id, revision)
        file_pattern = self._get_file_pattern(model_id, revision)

        # Delete artifact key
        await self._client.client.delete([artifact_key])

        # Find and delete all file keys
        # Note: SCAN is more efficient than KEYS for production use
        cursor = b"0"
        while True:
            result = await self._client.client.scan(cursor, match=file_pattern, count=100)
            if result is None or len(result) != 2:
                break

            cursor = cast(bytes, result[0])
            keys = cast(list[bytes], result[1])
            if keys:
                await self._client.client.delete(cast(list[str | bytes], keys))

            if cursor == b"0":
                break

        log.debug(
            "Cleaned up artifact download tracking: model_id={}, revision={}",
            model_id,
            revision,
        )



from pydantic import BaseModel, ConfigDict


class QueryTimeRange(BaseModel):
    """Time range parameters for a Prometheus query."""

    start: str
    end: str
    step: str

    model_config = ConfigDict(frozen=True)



"""
Common types for RBAC system.
"""

from __future__ import annotations

from enum import StrEnum

from ai.backend.common.data.permission.types import (
    EntityType,
    OperationType,
    PermissionStatus,
    RoleSource,
    RoleStatus,
)

__all__ = (
    "AssignedUserOrderField",
    "EntityType",
    "OperationType",
    "OrderDirection",
    "PermissionStatus",
    "RoleOrderField",
    "RoleSource",
    "RoleStatus",
    "ScopeOrderField",
)


class OrderDirection(StrEnum):
    """Order direction for sorting."""

    ASC = "asc"
    DESC = "desc"


class RoleOrderField(StrEnum):
    """Fields available for ordering roles."""

    NAME = "name"
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


class AssignedUserOrderField(StrEnum):
    """Fields available for ordering assigned users."""

    USERNAME = "username"
    EMAIL = "email"
    GRANTED_AT = "granted_at"


class ScopeOrderField(StrEnum):
    """Fields available for ordering scope IDs."""

    NAME = "name"
    CREATED_AT = "created_at"



from __future__ import annotations

import asyncio
import base64
import ipaddress
import numbers
import random
import re
import sys
import uuid
from collections import OrderedDict
from collections.abc import AsyncIterator, Iterable, Iterator, Mapping
from datetime import timedelta
from itertools import chain
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

import aiofiles
import yarl
from async_timeout import timeout
from pydantic import BaseModel

# It is a bad practice to keep all "miscellaneous" stuffs
# into the single "utils" module.
# Let's categorize them by purpose and domain, and keep
# refactoring to use the proper module names.
from .asyncio import (  # for legacy imports  # noqa
    AsyncBarrier,
    cancel_tasks,
    current_loop,
    run_through,
)
from .defs import DEFAULT_FILE_IO_TIMEOUT
from .enum_extension import StringSetFlag  # for legacy imports  # noqa
from .exception import (
    BaseNFSMountCheckFailed,
    ExportPathNotFound,
    NFSTimeoutError,
    NFSUnexpectedError,
    ShowmountFailed,
    ShowmountNotFound,
    VolumeMountFailed,
    VolumeUnmountFailed,
)
from .files import AsyncFileWriter  # for legacy imports  # noqa
from .networking import curl, find_free_port  # for legacy imports  # noqa
from .types import BinarySize

if TYPE_CHECKING:
    from decimal import Decimal

    from aiofiles.threadpool.text import AsyncTextIOWrapper

KT = TypeVar("KT")
VT = TypeVar("VT")


def env_info() -> str:
    """
    Returns a string that contains the Python version and runtime path.
    """
    v = sys.version_info
    pyver = f"Python {v.major}.{v.minor}.{v.micro}"
    if v.releaselevel == "alpha":
        pyver += "a"
    if v.releaselevel == "beta":
        pyver += "b"
    if v.releaselevel == "candidate":
        pyver += "rc"
    if v.releaselevel != "final":
        pyver += str(v.serial)
    return f"{pyver} (env: {sys.prefix})"


def odict[KT, VT](*args: tuple[KT, VT]) -> OrderedDict[KT, VT]:
    """
    A short-hand for the constructor of OrderedDict.
    :code:`odict(('a',1), ('b',2))` is equivalent to
    :code:`OrderedDict([('a',1), ('b',2)])`.
    """
    return OrderedDict(args)


def dict2kvlist[KT, VT](o: Mapping[KT, VT]) -> Iterable[KT | VT]:
    """
    Serializes a dict-like object into a generator of the flatten list of
    repeating key-value pairs.  It is useful when using HMSET method in Redis.

    Example:

       >>> list(dict2kvlist({'a': 1, 'b': 2}))
       ['a', 1, 'b', 2]
    """
    return chain.from_iterable((k, v) for k, v in o.items())


def generate_uuid() -> str:
    u = uuid.uuid4()
    # Strip the last two padding characters because u always has fixed length.
    return base64.urlsafe_b64encode(u.bytes)[:-2].decode("ascii")


def get_random_seq(length: float, num_points: int, min_distance: float) -> Iterator[float]:
    """
    Generate a random sequence of numbers within the range [0, length]
    with the given number of points and the minimum distance between the points.

    Note that X ( = the minimum distance d x the number of points N) must be equivalent to or smaller than
    the length L + d to guarantee the the minimum distance between the points.
    If X == L + d, the points are always equally spaced with d.

    :return: An iterator over the generated sequence
    """
    if num_points * min_distance > length + min_distance:
        raise ValueError(
            "There are too many points or it has a too large distance which cannot be fit into the"
            " given length."
        )
    extra = length - (num_points - 1) * min_distance
    ro = [random.uniform(0, 1) for _ in range(num_points + 1)]
    sum_ro = sum(ro)
    rn = [extra * r / sum_ro for r in ro[0:num_points]]
    spacing = [min_distance + rn[i] for i in range(num_points)]
    cumulative_sum = 0.0
    for s in spacing:
        cumulative_sum += s
        yield cumulative_sum - min_distance


def pprint_with_type(
    data: dict[Any, Any] | list[Any] | set[Any],
    depth: int = 0,
) -> None:
    """
    A pprint-like function that explicitly prints out the types of all objects.
    It is useful to debug JSON serialization errors due to hidden str-compatible keys
    in a nested dict.
    """
    indent = "  " * depth
    match data:
        case dict():  # dict
            for k, v in data.items():
                if isinstance(v, dict):
                    print(f"{indent}{k!r} ({type(k)=}): {{...: ...}}")
                    pprint_with_type(v, depth=depth + 1)
                    continue
                if isinstance(v, list):
                    print(f"{indent}{k!r} ({type(k)=}): [...]")
                    pprint_with_type(v, depth=depth + 1)
                    continue
                if isinstance(v, set):
                    print(f"{indent}{k!r} ({type(k)=}): {{...}}")
                    pprint_with_type(v, depth=depth + 1)
                    continue
                print(f"{indent}{k!r} ({type(k)=}): {v!r} ({type(v)=})")
        case list() | set():  # list, set
            for v in data:
                if isinstance(v, dict):
                    print(f"{indent}- {{...: ...}}")
                    pprint_with_type(v, depth=depth + 1)
                    continue
                if isinstance(v, list):
                    print(f"{indent}- [...]")
                    pprint_with_type(v, depth=depth + 1)
                    continue
                if isinstance(v, set):
                    print(f"{indent}- {{...}}")
                    pprint_with_type(v, depth=depth + 1)
                    continue
                print(f"{indent}- {v!r} ({type(v)=})")


def nmget(
    o: Mapping[str, Any],
    key_path: str,
    def_val: Any | None = None,
    path_delimiter: str = ".",
    null_as_default: bool = True,
) -> Any:
    """
    A short-hand for retrieving a value from nested mappings
    ("nested-mapping-get"). At each level it checks if the given "path"
    component in the given key exists and return the default value whenever
    fails.

    Example:
    >>> o = {'a':{'b':1}, 'x': None}
    >>> nmget(o, 'a', 0)
    {'b': 1}
    >>> nmget(o, 'a.b', 0)
    1
    >>> nmget(o, 'a/b', 0, '/')
    1
    >>> nmget(o, 'a.c', 0)
    0
    >>> nmget(o, 'x', 0)
    0
    >>> nmget(o, 'x', 0, null_as_default=False)
    None
    """
    pieces = key_path.split(path_delimiter)
    current: Any = o
    while pieces:
        p = pieces.pop(0)
        if current is None or p not in current:
            return def_val
        current = current[p]
    if current is None and null_as_default:
        return def_val
    return current


def readable_size_to_bytes(expr: Any) -> BinarySize | Decimal:
    if isinstance(expr, numbers.Integral):
        return BinarySize(expr)
    return BinarySize.from_str(expr)


def str_to_timedelta(tstr: str) -> timedelta:
    """
    Convert humanized timedelta string into a Python timedelta object.

    Example:
    >>> str_to_timedelta('30min')
    datetime.timedelta(seconds=1800)
    >>> str_to_timedelta('1d1hr')
    datetime.timedelta(days=1, seconds=3600)
    >>> str_to_timedelta('2hours 15min')
    datetime.timedelta(seconds=8100)
    >>> str_to_timedelta('20sec')
    datetime.timedelta(seconds=20)
    >>> str_to_timedelta('300')
    datetime.timedelta(seconds=300)
    >>> str_to_timedelta('-1day')
    datetime.timedelta(days=-1)
    """
    _rx = re.compile(
        r"(?P<sign>[+|-])?\s*"
        r"((?P<days>\d+(\.\d+)?)(d|day|days))?\s*"
        r"((?P<hours>\d+(\.\d+)?)(h|hr|hrs|hour|hours))?\s*"
        r"((?P<minutes>\d+(\.\d+)?)(m|min|mins|minute|minutes))?\s*"
        r"((?P<seconds>\d+(\.\d+)?)(s|sec|secs|second|seconds))?$"
    )
    match = _rx.match(tstr)
    if not match:
        try:
            return timedelta(seconds=float(tstr))  # consider bare number string as seconds
        except TypeError:
            pass
        raise ValueError("Invalid time expression")
    groups = match.groupdict()
    sign = groups.pop("sign", None)
    if set(groups.values()) == {None}:
        raise ValueError("Invalid time expression")
    params = {n: -float(t) if sign == "-" else float(t) for n, t in groups.items() if t}
    return timedelta(**params)


class FstabEntry:
    """
    Entry class represents a non-comment line on the `fstab` file.
    """

    def __init__(
        self, device: str, mountpoint: str, fstype: str, options: str | None, d: int = 0, p: int = 0
    ) -> None:
        self.device = device
        self.mountpoint = mountpoint
        self.fstype = fstype
        if not options:
            options = "defaults"
        self.options = options
        self.d = d
        self.p = p

    def __eq__(self, o: Any) -> bool:
        return str(self) == str(o)

    def __str__(self) -> str:
        return f"{self.device} {self.mountpoint} {self.fstype} {self.options} {self.d} {self.p}"


class Fstab:
    """
    Reader/writer for fstab file.
    Takes aiofile pointer for async I/O. It should be writable if add/remove
    operations are needed.

    NOTE: This class references Jorge Niedbalski R.'s gist snippet.
          We have been converted it to be compatible with Python 3
          and to support async I/O.
          (https://gist.github.com/niedbalski/507e974ed2d54a87ad37)
    """

    def __init__(self, fp: AsyncTextIOWrapper) -> None:
        self._fp = fp

    def _hydrate_entry(self, line: str) -> FstabEntry:
        parts = [x for x in line.strip("\n").split(" ") if x not in ("", None)]
        # Ensure we have at least 4 parts (device, mountpoint, fstype, options)
        # and convert the last two to integers if present
        device = parts[0] if len(parts) > 0 else ""
        mountpoint = parts[1] if len(parts) > 1 else ""
        fstype = parts[2] if len(parts) > 2 else ""
        options = parts[3] if len(parts) > 3 else None
        d = int(parts[4]) if len(parts) > 4 else 0
        p = int(parts[5]) if len(parts) > 5 else 0
        return FstabEntry(device, mountpoint, fstype, options, d, p)

    async def get_entries(self) -> AsyncIterator[FstabEntry]:
        await self._fp.seek(0)
        for line in await self._fp.readlines():
            try:
                line = line.strip()
                if not line.startswith("#"):
                    yield self._hydrate_entry(line)
            except TypeError:
                pass

    async def get_entry_by_attr(self, attr: str, value: Any) -> FstabEntry | None:
        async for entry in self.get_entries():
            e_attr = getattr(entry, attr)
            if e_attr == value:
                return entry
        return None

    async def add_entry(self, entry: FstabEntry) -> None:
        if await self.get_entry_by_attr("device", entry.device):
            return
        await self._fp.write(str(entry) + "\n")
        await self._fp.truncate()

    async def add(
        self,
        device: str,
        mountpoint: str,
        fstype: str,
        options: str | None = None,
        d: int = 0,
        p: int = 0,
    ) -> None:
        return await self.add_entry(FstabEntry(device, mountpoint, fstype, options, d, p))

    async def remove_entry(self, entry: FstabEntry) -> bool:
        await self._fp.seek(0)
        lines = await self._fp.readlines()
        line_no: int | None = None
        for index, line in enumerate(lines):
            if not line.strip().startswith("#"):
                try:
                    if self._hydrate_entry(line) == entry:
                        line_no = index
                        break
                except TypeError:
                    pass
        else:
            return False
        if line_no is None:
            raise RuntimeError("Failed to find entry in fstab")
        del lines[line_no]
        await self._fp.seek(0)
        await self._fp.write("".join(lines))
        await self._fp.truncate()
        return True

    async def remove_by_mountpoint(self, mountpoint: str) -> bool:
        entry = await self.get_entry_by_attr("mountpoint", mountpoint)
        if entry:
            return await self.remove_entry(entry)
        return False


async def check_nfs_remote_server(
    server: str,
    export_path: str,
) -> None:
    """
    Check if NFS export is available on the remote server.

    Args:
        server: NFS server hostname or IP address
        export_path: Export directory path to check (e.g., '/export')
    """
    try:
        # Execute showmount command to list exports
        process = await asyncio.create_subprocess_exec(
            "showmount",
            "-e",
            server,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        # Wait for command completion with timeout
        stdout, stderr = await asyncio.wait_for(process.communicate(), timeout=10.0)

        # Check if command succeeded
        if process.returncode != 0:
            error_msg = stderr.decode().strip()
            raise ShowmountFailed(f"showmount command failed: {error_msg}")

        # Parse output to find the export path
        output = stdout.decode().strip()
        lines = output.split("\n")

        # Skip header line (usually "Export list for server:")
        export_lines = [line for line in lines[1:] if line.strip()]

        # Check if the requested export path exists
        for line in export_lines:
            # Format is typically: /export  192.168.1.0/24
            # or: /export  *
            parts = line.split()
            if parts and parts[0] == export_path:
                # Export path found
                return

        # Export path not found
        raise ExportPathNotFound(f"Export '{export_path}' not found on server '{server}'.")
    except BaseNFSMountCheckFailed as e:
        raise e from None

    except TimeoutError as e:
        raise NFSTimeoutError(f"Timeout: No response from {server} within 10 seconds") from e

    except FileNotFoundError as e:
        raise ShowmountNotFound("showmount command not found. Install nfs-common package.") from e

    except Exception as e:
        raise NFSUnexpectedError(f"Unexpected error: {e!s}") from e


async def mount(
    mount_path: str,
    fs_location: str,
    fs_type: str = "nfs",
    cmd_options: str | None = None,
    edit_fstab: bool = False,
    fstab_path: str | None = None,
    mount_prefix: str | None = None,
) -> bool:
    if mount_prefix is None:
        mount_prefix = "/"
    if fstab_path is None:
        fstab_path = "/etc/fstab"
    mountpoint = Path(mount_prefix) / mount_path
    if mountpoint.is_mount():
        return False
    try:
        mountpoint.mkdir(exist_ok=True)
    except OSError as e:
        raise VolumeMountFailed(f"Failed to create mountpoint {mountpoint}. (e: {e})") from e
    if cmd_options is not None:
        cmd = [
            "mount",
            "-t",
            fs_type,
            "-o",
            cmd_options,
            fs_location,
            str(mountpoint),
        ]
    else:
        cmd = ["mount", "-t", fs_type, fs_location, str(mountpoint)]
    proc = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    raw_out, raw_err = await proc.communicate()
    raw_out.decode("utf8")
    err = raw_err.decode("utf8")
    await proc.wait()
    if err:
        raise VolumeMountFailed(f"Failed to mount {fs_location} on {mountpoint}. (e: {err})")
    if edit_fstab:
        async with aiofiles.open(fstab_path, mode="r+") as fp:
            fstab = Fstab(fp)
            await fstab.add(
                fs_location,
                str(mountpoint),
                fs_type,
                cmd_options,
            )
    return True


async def umount(
    mount_path: str,
    mount_prefix: str | None = None,
    edit_fstab: bool = False,
    fstab_path: str | None = None,
    rmdir_if_empty: bool = False,
    *,
    timeout_sec: float | None = DEFAULT_FILE_IO_TIMEOUT,
) -> bool:
    if mount_prefix is None:
        mount_prefix = "/"
    if fstab_path is None:
        fstab_path = "/etc/fstab"
    mountpoint = Path(mount_prefix) / mount_path
    if Path(mount_prefix) == mountpoint:
        raise ValueError("mountpoint cannot be the same as mount_prefix")
    if not mountpoint.is_mount():
        return False
    try:
        with timeout(timeout_sec):
            proc = await asyncio.create_subprocess_exec(
                *[
                    "umount",
                    str(mountpoint),
                ],
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            raw_out, raw_err = await proc.communicate()
            raw_out.decode("utf8")
            err = raw_err.decode("utf8")
            await proc.wait()
    except TimeoutError as e:
        raise VolumeUnmountFailed(
            f"Failed to umount {mountpoint}. Raise timeout ({timeout_sec}sec). "
            "The process may be hanging in state D, which needs to be checked."
        ) from e
    if err:
        raise VolumeUnmountFailed(f"Failed to umount {mountpoint}")
    if rmdir_if_empty:
        try:
            mountpoint.rmdir()  # delete directory if empty
        except OSError:
            pass
    if edit_fstab:
        async with aiofiles.open(fstab_path, mode="r+") as fp:
            fstab = Fstab(fp)
            await fstab.remove_by_mountpoint(str(mountpoint))
    return True


async def chown(path: Path | str, uid_gid: str, mount_prefix: str | Path | None = None) -> None:
    mounted = Path(mount_prefix or "/", path)
    await asyncio.create_subprocess_exec(
        "chown",
        uid_gid,
        str(mounted),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )


def is_ip_address_format(str: str) -> bool:
    try:
        url = yarl.URL("//" + str)
        return bool(url.host and ipaddress.ip_address(url.host))
    except ValueError:
        return False


def join_non_empty(*args: str | None, sep: str) -> str:
    """
    Joins non-empty strings from the given arguments using the specified separator.
    """
    filtered_args = [arg for arg in args if arg]
    return sep.join(filtered_args)


def b64encode(s: str) -> str:
    """
    base64 encoding method of graphql_relay.
    Use it in components where the graphql_relay package is unavailable.
    """
    b: bytes = s.encode("utf-8") if isinstance(s, str) else s
    return base64.b64encode(b).decode("ascii")


T = TypeVar("T", bound=BaseModel)


def _deep_merge_lists(base: list[Any], override: list[Any]) -> list[Any]:
    """
    Merge two lists by index. If both elements at the same index are Mappings,
    deep merge them. If both are lists, recursively merge them.
    Otherwise, override element takes precedence.
    """
    result: list[Any] = []
    max_len = max(len(base), len(override))
    for i in range(max_len):
        if i >= len(base):
            result.append(override[i])
        elif i >= len(override):
            result.append(base[i])
        elif isinstance(base[i], Mapping) and isinstance(override[i], Mapping):
            result.append(deep_merge(base[i], override[i]))
        elif isinstance(base[i], list) and isinstance(override[i], list):
            result.append(_deep_merge_lists(base[i], override[i]))
        elif override[i] is not None:
            result.append(override[i])
        else:
            result.append(base[i])
    return result


def deep_merge(*args: Mapping[str, Any]) -> Mapping[str, Any]:
    """
    Recursively merge any number of mappings.
    Later mappings override earlier ones on key conflicts.
    For lists, merge elements at the same index if both are Mappings.

    Example
    -------
    >>> deep_merge({"a": 1, "b": {"x": 1}},
    ...            {"b": {"y": 2}, "c": 3},
    ...            {"b": {"x": 10}})
    {'a': 1, 'b': {'x': 10, 'y': 2}, 'c': 3}

    >>> deep_merge({"items": [{"name": "a", "value": 1}]},
    ...            {"items": [{"name": "a", "extra": 2}]})
    {'items': [{'name': 'a', 'value': 1, 'extra': 2}]}
    """
    merged: dict[str, Any] = {}
    for m in args:
        for k, vb in m.items():
            va = merged.get(k)
            if isinstance(va, Mapping) and isinstance(vb, Mapping):
                merged[k] = deep_merge(va, vb)
            elif isinstance(va, list) and isinstance(vb, list):
                merged[k] = _deep_merge_lists(va, vb)
            elif vb is not None:
                merged[k] = vb
            else:
                merged[k] = va

    return merged


def addr_to_hostport_pair(addr: str) -> tuple[str, int]:
    """
    Convert a Redis address string to a host-port pair.
    """
    parts = addr.split(":")
    if len(parts) == 1:
        raise ValueError(f"Invalid Redis address format: {addr}")
    host = parts[0]
    port = int(parts[1])
    return host, port



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import NewType

# Service group is a string identifier with type safety
# Allows components to define their own service groups while maintaining type safety
ServiceGroup = NewType("ServiceGroup", str)

# Built-in service groups
MANAGER: ServiceGroup = ServiceGroup("manager")
AGENT: ServiceGroup = ServiceGroup("agent")
STORAGE_PROXY: ServiceGroup = ServiceGroup("storage-proxy")
APPPROXY: ServiceGroup = ServiceGroup("appproxy")
DATABASE: ServiceGroup = ServiceGroup("database")
ETCD: ServiceGroup = ServiceGroup("etcd")
REDIS: ServiceGroup = ServiceGroup("redis")
API: ServiceGroup = ServiceGroup("api")
CONTAINER: ServiceGroup = ServiceGroup("container")

# Component ID is a string identifier with type safety
# For services like database, redis, etcd: use service name (e.g., "postgres", "redis")
# For servers: use server-specific identifier
ComponentId = NewType("ComponentId", str)

# Built-in component IDs for database
CID_POSTGRES: ComponentId = ComponentId("postgres")

# Built-in component IDs for Redis/Valkey
CID_REDIS_ARTIFACT: ComponentId = ComponentId("artifact")
CID_REDIS_CONTAINER_LOG: ComponentId = ComponentId("container_log")
CID_REDIS_LIVE: ComponentId = ComponentId("live")
CID_REDIS_STAT: ComponentId = ComponentId("stat")
CID_REDIS_IMAGE: ComponentId = ComponentId("image")
CID_REDIS_STREAM: ComponentId = ComponentId("stream")
CID_REDIS_SCHEDULE: ComponentId = ComponentId("schedule")
CID_REDIS_BGTASK: ComponentId = ComponentId("bgtask")
CID_REDIS_SESSION: ComponentId = ComponentId("session")
CID_REDIS_CORE_LIVE: ComponentId = ComponentId("core_live")

# Built-in component IDs for etcd
CID_ETCD: ComponentId = ComponentId("etcd")

# Built-in component IDs for container
CID_DOCKER: ComponentId = ComponentId("docker")


@dataclass(frozen=True)
class HealthCheckKey:
    """
    Unique key to identify a health checker registration.

    Uses frozen dataclass to be hashable and usable as dict key.
    """

    service_group: ServiceGroup
    component_id: ComponentId


@dataclass
class ComponentHealthStatus:
    """
    Health status for a single component.

    Represents the result of checking one component within a service group.
    """

    is_healthy: bool
    last_checked_at: datetime
    error_message: str | None = None


@dataclass
class ServiceHealth:
    """
    Health status of a service group containing multiple components.

    A service health checker checks multiple components and returns
    their individual statuses aggregated in this structure.
    """

    results: dict[ComponentId, ComponentHealthStatus]


@dataclass
class AllServicesHealth:
    """
    Aggregated health status of all service groups.

    Contains health results from all registered service health checkers.
    """

    results: dict[ServiceGroup, ServiceHealth]



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class MappingRule:
    """A rule that maps an existing TOML config field to a service-discovery endpoint."""

    source: str
    role: str
    scope: str
    protocol: str
    fallback: str | None = None


@dataclass
class DetectedField:
    """A field detected from the existing config file."""

    key_path: str
    host: str
    port: int
    is_fallback: bool


@dataclass
class GeneratedEndpoint:
    """An endpoint generated from a detected field."""

    role: str
    scope: str
    address: str
    port: int
    protocol: str
    source_field: DetectedField


@dataclass
class MigrationResult:
    """The result of a config migration run."""

    detected_fields: list[DetectedField]
    generated_endpoints: list[GeneratedEndpoint]
    skipped_rules: list[tuple[MappingRule, str]]
    has_existing_endpoints: bool



from .service_discovery import (
    MODEL_SERVICE_GROUP,
    MODEL_SERVICE_ROUTE_TTL,
    HealthStatus,
    ModelServiceMetadata,
    ServiceDiscovery,
    ServiceDiscoveryLoop,
    ServiceEndpoint,
    ServiceMetadata,
)

__all__ = [
    "MODEL_SERVICE_GROUP",
    "MODEL_SERVICE_ROUTE_TTL",
    "HealthStatus",
    "ModelServiceMetadata",
    "ServiceDiscovery",
    "ServiceDiscoveryLoop",
    "ServiceEndpoint",
    "ServiceMetadata",
]



"""
Response DTOs for notification system.
Shared between Client SDK and Manager API.
"""

from __future__ import annotations

from datetime import datetime
from typing import Any
from uuid import UUID

from pydantic import BaseModel, Field

from ai.backend.common.api_handlers import BaseResponseModel
from ai.backend.common.data.notification.types import (
    EmailMessage,
    NotificationChannelType,
    NotificationRuleType,
    SMTPAuth,
    SMTPConnection,
)

__all__ = (
    "CreateNotificationChannelResponse",
    "CreateNotificationRuleResponse",
    "DeleteNotificationChannelResponse",
    "DeleteNotificationRuleResponse",
    "GetNotificationChannelResponse",
    "GetNotificationRuleResponse",
    "ListNotificationChannelsResponse",
    "ListNotificationRuleTypesResponse",
    "ListNotificationRulesResponse",
    "NotificationChannelDTO",
    "NotificationRuleDTO",
    "NotificationRuleTypeSchemaResponse",
    "PaginationInfo",
    "UpdateNotificationChannelResponse",
    "UpdateNotificationRuleResponse",
    "ValidateNotificationChannelResponse",
    "ValidateNotificationRuleResponse",
    "WebhookSpecResponse",
)


class WebhookSpecResponse(BaseResponseModel):
    """Response model for webhook specification."""

    url: str = Field(description="Webhook URL")


class EmailSpecResponse(BaseResponseModel):
    """Response model for email specification."""

    smtp: SMTPConnection = Field(description="SMTP connection settings")
    message: EmailMessage = Field(description="Email message settings")
    auth: SMTPAuth | None = Field(default=None, description="SMTP authentication credentials")


class NotificationChannelDTO(BaseModel):
    """DTO for notification channel data."""

    id: UUID = Field(description="Channel ID")
    name: str = Field(description="Channel name")
    description: str | None = Field(default=None, description="Channel description")
    channel_type: NotificationChannelType = Field(description="Channel type")
    spec: WebhookSpecResponse | EmailSpecResponse = Field(description="Channel configuration")
    enabled: bool = Field(description="Whether the channel is enabled")
    created_at: datetime = Field(description="Creation timestamp")
    created_by: UUID = Field(description="ID of user who created the channel")
    updated_at: datetime = Field(description="Last update timestamp")


class NotificationRuleDTO(BaseModel):
    """DTO for notification rule data."""

    id: UUID = Field(description="Rule ID")
    name: str = Field(description="Rule name")
    description: str | None = Field(default=None, description="Rule description")
    rule_type: NotificationRuleType = Field(description="Rule type")
    channel: NotificationChannelDTO = Field(description="Associated channel")
    message_template: str = Field(description="Jinja2 template for notification message")
    enabled: bool = Field(description="Whether the rule is enabled")
    created_at: datetime = Field(description="Creation timestamp")
    created_by: UUID = Field(description="ID of user who created the rule")
    updated_at: datetime = Field(description="Last update timestamp")


class CreateNotificationChannelResponse(BaseResponseModel):
    """Response for creating a notification channel."""

    channel: NotificationChannelDTO = Field(description="Created channel")


class UpdateNotificationChannelResponse(BaseResponseModel):
    """Response for updating a notification channel."""

    channel: NotificationChannelDTO = Field(description="Updated channel")


class DeleteNotificationChannelResponse(BaseResponseModel):
    """Response for deleting a notification channel."""

    deleted: bool = Field(description="Whether the channel was deleted")


class GetNotificationChannelResponse(BaseResponseModel):
    """Response for getting a notification channel."""

    channel: NotificationChannelDTO = Field(description="Channel data")


class PaginationInfo(BaseModel):
    """Pagination information."""

    total: int = Field(description="Total number of items")
    offset: int = Field(description="Number of items skipped")
    limit: int | None = Field(default=None, description="Maximum items returned")


class ListNotificationChannelsResponse(BaseResponseModel):
    """Response for listing notification channels."""

    channels: list[NotificationChannelDTO] = Field(description="List of channels")
    pagination: PaginationInfo = Field(description="Pagination information")


class CreateNotificationRuleResponse(BaseResponseModel):
    """Response for creating a notification rule."""

    rule: NotificationRuleDTO = Field(description="Created rule")


class UpdateNotificationRuleResponse(BaseResponseModel):
    """Response for updating a notification rule."""

    rule: NotificationRuleDTO = Field(description="Updated rule")


class DeleteNotificationRuleResponse(BaseResponseModel):
    """Response for deleting a notification rule."""

    deleted: bool = Field(description="Whether the rule was deleted")


class GetNotificationRuleResponse(BaseResponseModel):
    """Response for getting a notification rule."""

    rule: NotificationRuleDTO = Field(description="Rule data")


class ListNotificationRulesResponse(BaseResponseModel):
    """Response for listing notification rules."""

    rules: list[NotificationRuleDTO] = Field(description="List of rules")
    pagination: PaginationInfo = Field(description="Pagination information")


class ValidateNotificationChannelResponse(BaseResponseModel):
    """Response for validating a notification channel."""

    channel_id: UUID = Field(description="ID of the validated channel")


class ValidateNotificationRuleResponse(BaseResponseModel):
    """Response for validating a notification rule."""

    message: str = Field(description="The rendered message from the template")


class ListNotificationRuleTypesResponse(BaseResponseModel):
    """Response for listing available notification rule types."""

    rule_types: list[NotificationRuleType] = Field(
        description="List of available notification rule types"
    )


class NotificationRuleTypeSchemaResponse(BaseResponseModel):
    """Response for getting notification rule type schema."""

    rule_type: NotificationRuleType = Field(
        description="The notification rule type for which the schema is provided"
    )
    json_schema: dict[str, Any] = Field(
        description="JSON schema describing the required notification_data structure for this rule type"
    )



"""
Request DTOs for auth system.
Shared between Client SDK and Manager API.
"""

from __future__ import annotations

from uuid import UUID

from pydantic import AliasChoices, Field

from ai.backend.common.api_handlers import BaseRequestModel

from .types import AuthTokenType

__all__ = (
    "AuthorizeRequest",
    "GetRoleRequest",
    "SignupRequest",
    "SignoutRequest",
    "UpdateFullNameRequest",
    "UpdatePasswordRequest",
    "UpdatePasswordNoAuthRequest",
    "UploadSSHKeypairRequest",
    "VerifyAuthRequest",
)


class AuthorizeRequest(BaseRequestModel):
    """Request to authorize a user and obtain API credentials."""

    type: AuthTokenType = Field(
        description="Token type to issue: 'keypair' for HMAC auth or 'jwt' for stateless auth",
    )
    domain: str = Field(
        description="Domain (tenant) the user belongs to; users are scoped per domain",
    )
    username: str = Field(
        description="Login identifier (email address or username)",
    )
    password: str = Field(
        description="Login password, verified against the stored hash",
    )
    stoken: str | None = Field(
        default=None,
        description="Secondary token forwarded to auth hook plugins (e.g., for 2FA)",
        validation_alias=AliasChoices("stoken", "sToken"),
    )


class GetRoleRequest(BaseRequestModel):
    """Request to get the user's roles at global, domain, and optionally group scope."""

    group: UUID | None = Field(
        default=None,
        description="Group (project) ID to also query group-level role; errors if not a member",
    )


class SignupRequest(BaseRequestModel):
    """Request to sign up a new user. Created with INACTIVE status; requires separate activation."""

    domain: str = Field(description="Domain (tenant) to create the user in")
    email: str = Field(
        description="Email for the account; must be globally unique, also used as username if not specified",
    )
    password: str = Field(
        description="Initial password; validated via VERIFY_PASSWORD_FORMAT hook",
    )
    username: str | None = Field(
        default=None,
        description="Display username; defaults to email if omitted",
    )
    full_name: str | None = Field(
        default=None,
        description="User's full name for display; defaults to empty string",
    )
    description: str | None = Field(
        default=None,
        description="Free-text account description; defaults to empty string",
    )


class SignoutRequest(BaseRequestModel):
    """Request to deactivate a user account (sets INACTIVE, does not delete)."""

    email: str = Field(
        description="Email of the account to deactivate; must match the requester's own email",
        validation_alias=AliasChoices("email", "username"),
    )
    password: str = Field(description="Current password for identity confirmation")


class UpdateFullNameRequest(BaseRequestModel):
    """Request to update the authenticated user's display full name."""

    email: str = Field(description="Email of the user to update")
    full_name: str = Field(description="New full name to set")


class UpdatePasswordRequest(BaseRequestModel):
    """Request to update the authenticated user's password with confirmation."""

    old_password: str = Field(description="Current password for identity verification")
    new_password: str = Field(
        description="Desired new password; validated via VERIFY_PASSWORD_FORMAT hook"
    )
    new_password2: str = Field(
        description="New password confirmation; must match new_password exactly"
    )


class UpdatePasswordNoAuthRequest(BaseRequestModel):
    """Request to update an expired password without API auth.
    Only available when max_password_age is configured."""

    domain: str = Field(description="Domain (tenant) the user belongs to")
    username: str = Field(description="Login identifier (email or username)")
    current_password: str = Field(description="Current (expired) password for verification")
    new_password: str = Field(description="New password; must differ from current password")


class UploadSSHKeypairRequest(BaseRequestModel):
    """Request to upload an SSH keypair. The pair is validated for consistency."""

    pubkey: str = Field(description="SSH public key in standard format (e.g., 'ssh-rsa AAAA...')")
    privkey: str = Field(description="SSH private key in PEM format")


class VerifyAuthRequest(BaseRequestModel):
    """Request to verify that the current API credentials are valid."""

    echo: str = Field(description="Arbitrary string echoed back to confirm auth is working")



import hashlib
import hmac
from collections.abc import Mapping
from datetime import datetime

from yarl import URL


def generate_signature(
    *,
    method: str,
    version: str,
    endpoint: URL,
    date: datetime,
    rel_url: str,
    content_type: str,
    access_key: str,
    secret_key: str,
    hash_type: str,
) -> tuple[Mapping[str, str], str]:
    """
    Generates the API request signature from the given parameters.
    """
    hash_type = hash_type
    hostname = endpoint.raw_authority
    body_hash = hashlib.new(hash_type, b"").hexdigest()

    sign_str = f"{method.upper()}\n{rel_url}\n{date.isoformat()}\nhost:{hostname}\ncontent-type:{content_type.lower()}\nx-backendai-version:{version}\n{body_hash}"
    sign_bytes = sign_str.encode()

    sign_key = hmac.new(secret_key.encode(), date.strftime("%Y%m%d").encode(), hash_type).digest()
    sign_key = hmac.new(sign_key, hostname.encode(), hash_type).digest()

    signature = hmac.new(sign_key, sign_bytes, hash_type).hexdigest()
    headers = {
        "Authorization": f"BackendAI signMethod=HMAC-{hash_type.upper()}, credential={access_key}:{signature}",
    }
    return headers, signature



from __future__ import annotations

import enum
from typing import Any

__all__ = ("StringSetFlag",)


class StringSetFlag(enum.StrEnum):
    def __eq__(self, other: Any) -> bool:
        result: bool = self.value == other
        return result

    def __hash__(self) -> int:
        return hash(self.value)

    def __or__(self, other: Any) -> set[str]:
        if isinstance(other, type(self)):
            other = other.value
        if not isinstance(other, (set, frozenset)):
            other = {other}
        result: set[str] = {self.value} | other
        return result

    __ror__ = __or__

    def __and__(self, other: Any) -> bool:
        if isinstance(other, (set, frozenset)):
            return self.value in other
        if isinstance(other, str):
            return self.value == other
        raise TypeError

    __rand__ = __and__

    def __xor__(self, other: Any) -> set[str] | str:
        if isinstance(other, (set, frozenset)):
            return {self.value} ^ other
        if isinstance(other, str):
            if other == self.value:
                return set()
            return other
        raise TypeError

    def __rxor__(self, other: Any) -> set[str] | str:
        if isinstance(other, (set, frozenset)):
            return set(other) ^ {self.value}
        if isinstance(other, str):
            if other == self.value:
                return set()
            return other
        raise TypeError

    def __str__(self) -> str:
        return self.value



from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import Any, Self, override

from ai.backend.common.data.agent.types import AgentInfo
from ai.backend.common.data.image.types import ScannedImage
from ai.backend.common.events.types import (
    AbstractAnycastEvent,
    EventDomain,
)
from ai.backend.common.events.user_event.user_event import UserEvent
from ai.backend.common.types import (
    AgentId,
    ImageCanonical,
)
from ai.backend.logging.types import LogLevel


class BaseAgentEvent(AbstractAnycastEvent):
    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.AGENT


@dataclass
class BaseAgentLifecycleEvent(BaseAgentEvent):
    reason: str

    @override
    def serialize(self) -> tuple[Any, ...]:
        return (self.reason,)

    @classmethod
    @override
    def deserialize(cls, value: tuple[Any, ...]) -> Self:
        return cls(value[0])

    @override
    def domain_id(self) -> str | None:
        return None

    @override
    def user_event(self) -> UserEvent | None:
        return None


@dataclass
class AgentStartedEvent(BaseAgentLifecycleEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "agent_started"


@dataclass
class AgentTerminatedEvent(BaseAgentLifecycleEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "agent_terminated"


@dataclass
class AgentOperationEvent(BaseAgentEvent):
    @override
    def domain_id(self) -> str | None:
        return None

    @override
    def user_event(self) -> UserEvent | None:
        return None


@dataclass
class AgentErrorEvent(AgentOperationEvent):
    message: str
    traceback: str | None = None
    user: Any | None = None
    context_env: Mapping[str, Any] = field(default_factory=dict)
    severity: LogLevel = LogLevel.ERROR

    @override
    def serialize(self) -> tuple[Any, ...]:
        return (
            self.message,
            self.traceback,
            self.user,
            self.context_env,
            self.severity.value,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple[Any, ...]) -> Self:
        return cls(
            value[0],
            value[1],
            value[2],
            value[3],
            LogLevel(value[4]),
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "agent_error"


@dataclass
class AgentHeartbeatEvent(AgentOperationEvent):
    agent_info: AgentInfo

    @override
    def serialize(self) -> tuple[Any, ...]:
        return (self.agent_info.model_dump(),)

    @classmethod
    @override
    def deserialize(cls, value: tuple[Any, ...]) -> Self:
        return cls(AgentInfo.model_validate(value[0]))

    @classmethod
    @override
    def event_name(cls) -> str:
        return "agent_heartbeat"


# For compatibility with redis key made with image canonical strings
# Use AgentInstalledImagesRemoveEvent instead of this if possible
@dataclass
class AgentImagesRemoveEvent(AgentOperationEvent):
    image_canonicals: list[ImageCanonical]

    @override
    def serialize(self) -> tuple[Any, ...]:
        return (self.image_canonicals,)

    @classmethod
    @override
    def deserialize(cls, value: tuple[Any, ...]) -> Self:
        return cls(value[0])

    @classmethod
    @override
    def event_name(cls) -> str:
        return "agent_images_remove"


@dataclass
class AgentInstalledImagesRemoveEvent(AgentOperationEvent):
    scanned_images: Mapping[ImageCanonical, ScannedImage]

    @override
    def serialize(self) -> tuple[Any, ...]:
        result = {}
        for canonical, image in self.scanned_images.items():
            result[str(canonical)] = image.to_dict()
        return (result,)

    @classmethod
    @override
    def deserialize(cls, value: tuple[Any, ...]) -> Self:
        result = {}
        for canonical, image_data in value[0].items():
            result[ImageCanonical(canonical)] = ScannedImage.from_dict(image_data)
        return cls(result)

    @classmethod
    @override
    def event_name(cls) -> str:
        return "agent_installed_images_remove"


@dataclass
class DoAgentResourceCheckEvent(AgentOperationEvent):
    agent_id: AgentId

    @override
    def serialize(self) -> tuple[Any, ...]:
        return (self.agent_id,)

    @classmethod
    @override
    def deserialize(cls, value: tuple[Any, ...]) -> Self:
        return cls(
            AgentId(value[0]),
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "do_agent_resource_check"



from __future__ import annotations

import asyncio
from datetime import UTC, datetime
from typing import Protocol, runtime_checkable

from ai.backend.common.health_checker.abc import StaticServiceHealthChecker
from ai.backend.common.health_checker.types import (
    REDIS,
    ComponentHealthStatus,
    ComponentId,
    ServiceGroup,
    ServiceHealth,
)


@runtime_checkable
class ValkeyPingable(Protocol):
    """Protocol for any client that supports ping() method."""

    async def ping(self) -> None:
        """Ping the server to check connection health."""
        ...


class ValkeyHealthChecker(StaticServiceHealthChecker):
    """
    Health checker for Valkey/Redis connections.

    Checks multiple Valkey/Redis clients and returns health status for each component.
    Each client should implement the ValkeyPingable protocol (i.e., has a ping() method).
    """

    _clients: dict[ComponentId, ValkeyPingable]
    _timeout: float
    _component_timeout: float

    def __init__(
        self,
        clients: dict[ComponentId, ValkeyPingable],
        timeout: float = 5.0,
        component_timeout: float = 2.0,
    ) -> None:
        """
        Initialize ValkeyHealthChecker.

        Args:
            clients: Dictionary mapping ComponentId to client instances
                    (e.g., {ComponentId("artifact"): client1, ComponentId("live"): client2})
            timeout: Timeout in seconds for the entire health check (service group timeout)
            component_timeout: Timeout in seconds for each component's ping (should be < timeout)
        """
        self._clients = clients
        self._timeout = timeout
        self._component_timeout = component_timeout

    @property
    def target_service_group(self) -> ServiceGroup:
        """The service group this checker monitors."""
        return REDIS

    async def _check_component(
        self,
        component_id: ComponentId,
        client: ValkeyPingable,
        check_time: datetime,
    ) -> tuple[ComponentId, ComponentHealthStatus]:
        """
        Check health of a single Valkey/Redis component with timeout.

        Args:
            component_id: The component identifier
            client: The client to check
            check_time: The timestamp for the check

        Returns:
            Tuple of (component_id, status)
        """
        try:
            async with asyncio.timeout(self._component_timeout):
                await client.ping()
            status = ComponentHealthStatus(
                is_healthy=True,
                last_checked_at=check_time,
                error_message=None,
            )
        except TimeoutError:
            status = ComponentHealthStatus(
                is_healthy=False,
                last_checked_at=check_time,
                error_message=f"Component timeout after {self._component_timeout}s",
            )
        except Exception as e:
            status = ComponentHealthStatus(
                is_healthy=False,
                last_checked_at=check_time,
                error_message=str(e),
            )

        return (component_id, status)

    async def check_service(self) -> ServiceHealth:
        """
        Check health of all Valkey/Redis clients concurrently.

        Returns:
            ServiceHealth containing status for each component
        """
        check_time = datetime.now(UTC)

        # Run all health checks concurrently
        tasks = [
            self._check_component(component_id, client, check_time)
            for component_id, client in self._clients.items()
        ]

        check_results = await asyncio.gather(*tasks, return_exceptions=True)

        # Process results
        results: dict[ComponentId, ComponentHealthStatus] = {}
        for i, result in enumerate(check_results):
            if isinstance(result, BaseException):
                # This shouldn't happen since _check_component catches all exceptions,
                # but handle it just in case
                component_id = list(self._clients.keys())[i]
                results[component_id] = ComponentHealthStatus(
                    is_healthy=False,
                    last_checked_at=check_time,
                    error_message=f"Unexpected error: {result}",
                )
            else:
                comp_id, status = result
                results[comp_id] = status

        return ServiceHealth(results=results)

    @property
    def timeout(self) -> float:
        """The timeout for the entire health check in seconds."""
        return self._timeout



from __future__ import annotations

from contextlib import AsyncExitStack
from typing import TYPE_CHECKING, Any

from ai.backend.common.dependencies.base import DependencyStack

if TYPE_CHECKING:
    from ai.backend.common.dependencies.base import (
        DependencyComposer,
        DependencyProvider,
        ResourcesT,
        ResourceT,
        SetupInputT,
    )
    from ai.backend.common.health_checker import ServiceHealthChecker
    from ai.backend.common.health_checker.types import ServiceGroup


class DependencyBuilderStack(DependencyStack):
    """
    DependencyStack that collects health checkers from providers.

    Uses AsyncExitStack internally for lifecycle management while automatically
    collecting health checkers from providers that implement gen_health_checkers() method.

    Health checkers are registered by their ServiceGroup (e.g., REDIS, DATABASE, ETCD).
    Each ServiceGroup can have only one health checker, which checks multiple components
    within that service group.

    Health checkers can be retrieved after dependency initialization
    to register them with a HealthProbe for health monitoring.
    """

    _stack: AsyncExitStack
    _health_checkers: dict[ServiceGroup, ServiceHealthChecker]

    def __init__(self) -> None:
        self._stack = AsyncExitStack()
        self._health_checkers = {}

    async def enter_dependency(
        self,
        provider: DependencyProvider[SetupInputT, ResourceT],
        setup_input: SetupInputT,
    ) -> ResourceT:
        """
        Execute a dependency provider and collect health checker.

        If the provider returns a health checker, it will be registered using
        checker.target_service_group as the key.
        """
        resource = await self._stack.enter_async_context(provider.provide(setup_input))

        # Collect health checker from provider
        checker = provider.gen_health_checkers(resource)
        if checker is not None:
            self._health_checkers[checker.target_service_group] = checker

        return resource

    async def enter_composer(
        self,
        composer: DependencyComposer[SetupInputT, ResourcesT],
        setup_input: SetupInputT,
    ) -> ResourcesT:
        """
        Execute a dependency composer and collect health checkers from nested dependencies.

        Creates a nested DependencyBuilderStack to track dependencies
        within the composer, then merges collected health checkers.
        """
        # Create nested builder stack
        nested_stack = DependencyBuilderStack()
        await self._stack.enter_async_context(nested_stack)

        # Compose and get resources
        resources = await nested_stack._stack.enter_async_context(
            composer.compose(nested_stack, setup_input)
        )

        # Collect health checkers from nested stack
        for service_group, checker in nested_stack._health_checkers.items():
            self._health_checkers[service_group] = checker

        return resources

    def get_health_checkers(self) -> dict[ServiceGroup, ServiceHealthChecker]:
        """
        Return collected health checkers.

        Returns:
            Dictionary mapping ServiceGroup to ServiceHealthChecker instances
            collected from all initialized dependencies.
        """
        return self._health_checkers.copy()

    async def __aenter__(self) -> DependencyBuilderStack:
        """
        Enter the async context.
        """
        await self._stack.__aenter__()
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any
    ) -> bool | None:
        """
        Exit the async context and cleanup resources in LIFO order.
        """
        return await self._stack.__aexit__(exc_type, exc_val, exc_tb)



from __future__ import annotations

from enum import StrEnum

from pydantic import Field

from ai.backend.common.api_handlers import BaseRequestModel
from ai.backend.common.dto.manager.query import StringFilter, UUIDFilter

__all__ = (
    "OrderDirection",
    "SchedulingResultType",
    "SessionHistoryOrderField",
    "SessionHistoryFilter",
    "SessionHistoryOrder",
    "DeploymentHistoryOrderField",
    "DeploymentHistoryFilter",
    "DeploymentHistoryOrder",
    "RouteHistoryOrderField",
    "RouteHistoryFilter",
    "RouteHistoryOrder",
)


class OrderDirection(StrEnum):
    ASC = "asc"
    DESC = "desc"


class SchedulingResultType(StrEnum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    STALE = "STALE"


class SessionHistoryOrderField(StrEnum):
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


class SessionHistoryFilter(BaseRequestModel):
    """Filter for session scheduling history."""

    session_id: UUIDFilter | None = Field(default=None, description="Filter by session ID")
    phase: StringFilter | None = Field(default=None, description="Filter by phase (contains)")
    from_status: list[str] | None = Field(default=None, description="Filter by from_status")
    to_status: list[str] | None = Field(default=None, description="Filter by to_status")
    result: list[SchedulingResultType] | None = Field(
        default=None, description="Filter by result (SUCCESS, FAILURE, STALE)"
    )
    error_code: StringFilter | None = Field(default=None, description="Filter by error code")
    message: StringFilter | None = Field(default=None, description="Filter by message (contains)")


class SessionHistoryOrder(BaseRequestModel):
    """Order specification for session scheduling history."""

    field: SessionHistoryOrderField = Field(description="Field to order by")
    direction: OrderDirection = Field(default=OrderDirection.DESC, description="Order direction")


class DeploymentHistoryOrderField(StrEnum):
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


class DeploymentHistoryFilter(BaseRequestModel):
    """Filter for deployment history."""

    deployment_id: UUIDFilter | None = Field(default=None, description="Filter by deployment ID")
    phase: StringFilter | None = Field(default=None, description="Filter by phase (contains)")
    from_status: list[str] | None = Field(default=None, description="Filter by from_status")
    to_status: list[str] | None = Field(default=None, description="Filter by to_status")
    result: list[SchedulingResultType] | None = Field(default=None, description="Filter by result")
    error_code: StringFilter | None = Field(default=None, description="Filter by error code")
    message: StringFilter | None = Field(default=None, description="Filter by message (contains)")


class DeploymentHistoryOrder(BaseRequestModel):
    """Order specification for deployment history."""

    field: DeploymentHistoryOrderField = Field(description="Field to order by")
    direction: OrderDirection = Field(default=OrderDirection.DESC, description="Order direction")


class RouteHistoryOrderField(StrEnum):
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


class RouteHistoryFilter(BaseRequestModel):
    """Filter for route history."""

    route_id: UUIDFilter | None = Field(default=None, description="Filter by route ID")
    deployment_id: UUIDFilter | None = Field(default=None, description="Filter by deployment ID")
    phase: StringFilter | None = Field(default=None, description="Filter by phase (contains)")
    from_status: list[str] | None = Field(default=None, description="Filter by from_status")
    to_status: list[str] | None = Field(default=None, description="Filter by to_status")
    result: list[SchedulingResultType] | None = Field(default=None, description="Filter by result")
    error_code: StringFilter | None = Field(default=None, description="Filter by error code")
    message: StringFilter | None = Field(default=None, description="Filter by message (contains)")


class RouteHistoryOrder(BaseRequestModel):
    """Order specification for route history."""

    field: RouteHistoryOrderField = Field(description="Field to order by")
    direction: OrderDirection = Field(default=OrderDirection.DESC, description="Order direction")



"""Request DTOs for Fair Share API."""

from __future__ import annotations

from decimal import Decimal
from uuid import UUID

from pydantic import BaseModel, Field

from ai.backend.common.api_handlers import BaseRequestModel

from .types import (
    DomainFairShareFilter,
    DomainFairShareOrder,
    DomainUsageBucketFilter,
    DomainUsageBucketOrder,
    ProjectFairShareFilter,
    ProjectFairShareOrder,
    ProjectUsageBucketFilter,
    ProjectUsageBucketOrder,
    UserFairShareFilter,
    UserFairShareOrder,
    UserUsageBucketFilter,
    UserUsageBucketOrder,
)

__all__ = (
    # Path parameters - Get
    "GetDomainFairSharePathParam",
    "GetProjectFairSharePathParam",
    "GetUserFairSharePathParam",
    # Path parameters - RG Scoped Get
    "RGDomainFairSharePathParam",
    "RGProjectFairSharePathParam",
    "RGUserFairSharePathParam",
    # Path parameters - RG Scoped Search
    "RGDomainFairShareSearchPathParam",
    "RGProjectFairShareSearchPathParam",
    "RGUserFairShareSearchPathParam",
    # Path parameters - RG Scoped Usage Bucket Search
    "RGDomainUsageBucketSearchPathParam",
    "RGProjectUsageBucketSearchPathParam",
    "RGUserUsageBucketSearchPathParam",
    # Path parameters - Upsert Weight
    "UpsertDomainFairShareWeightPathParam",
    "UpsertProjectFairShareWeightPathParam",
    "UpsertUserFairShareWeightPathParam",
    # Path parameters - Update Spec
    "UpdateResourceGroupFairShareSpecPathParam",
    # Get requests (deprecated, use PathParam)
    "GetDomainFairShareRequest",
    "GetProjectFairShareRequest",
    "GetUserFairShareRequest",
    # Search requests
    "SearchDomainFairSharesRequest",
    "SearchProjectFairSharesRequest",
    "SearchUserFairSharesRequest",
    "SearchDomainUsageBucketsRequest",
    "SearchProjectUsageBucketsRequest",
    "SearchUserUsageBucketsRequest",
    # Upsert Weight Requests
    "UpsertDomainFairShareWeightRequest",
    "UpsertProjectFairShareWeightRequest",
    "UpsertUserFairShareWeightRequest",
    # Bulk Upsert Weight Requests
    "DomainWeightEntryInput",
    "BulkUpsertDomainFairShareWeightRequest",
    "ProjectWeightEntryInput",
    "BulkUpsertProjectFairShareWeightRequest",
    "UserWeightEntryInput",
    "BulkUpsertUserFairShareWeightRequest",
    # Update Spec Requests
    "ResourceWeightEntryInput",
    "UpdateResourceGroupFairShareSpecRequest",
)


# Path Parameters


class GetDomainFairSharePathParam(BaseRequestModel):
    """Path parameters for getting a single domain fair share."""

    resource_group: str = Field(description="Scaling group name")
    domain_name: str = Field(description="Domain name")


class GetProjectFairSharePathParam(BaseRequestModel):
    """Path parameters for getting a single project fair share."""

    resource_group: str = Field(description="Scaling group name")
    project_id: UUID = Field(description="Project ID")


class GetUserFairSharePathParam(BaseRequestModel):
    """Path parameters for getting a single user fair share."""

    resource_group: str = Field(description="Scaling group name")
    project_id: UUID = Field(description="Project ID")
    user_uuid: UUID = Field(description="User UUID")


# RG-Scoped Path Parameters


class RGDomainFairSharePathParam(BaseRequestModel):
    """Path parameters for RG-scoped domain fair share get."""

    resource_group: str = Field(description="Scaling group name")
    domain_name: str = Field(description="Domain name")


class RGProjectFairSharePathParam(BaseRequestModel):
    """Path parameters for RG-scoped project fair share get."""

    resource_group: str = Field(description="Scaling group name")
    domain_name: str = Field(description="Domain name")
    project_id: UUID = Field(description="Project ID")


class RGUserFairSharePathParam(BaseRequestModel):
    """Path parameters for RG-scoped user fair share get."""

    resource_group: str = Field(description="Scaling group name")
    domain_name: str = Field(description="Domain name")
    project_id: UUID = Field(description="Project ID")
    user_uuid: UUID = Field(description="User UUID")


class RGDomainFairShareSearchPathParam(BaseRequestModel):
    """Path parameters for RG-scoped domain fair share search."""

    resource_group: str = Field(description="Scaling group name")


class RGProjectFairShareSearchPathParam(BaseRequestModel):
    """Path parameters for RG-scoped project fair share search."""

    resource_group: str = Field(description="Scaling group name")
    domain_name: str = Field(description="Domain name")


class RGUserFairShareSearchPathParam(BaseRequestModel):
    """Path parameters for RG-scoped user fair share search."""

    resource_group: str = Field(description="Scaling group name")
    domain_name: str = Field(description="Domain name")
    project_id: UUID = Field(description="Project ID")


class RGDomainUsageBucketSearchPathParam(BaseRequestModel):
    """Path parameters for RG-scoped domain usage bucket search."""

    resource_group: str = Field(description="Resource group name")


class RGProjectUsageBucketSearchPathParam(BaseRequestModel):
    """Path parameters for RG-scoped project usage bucket search."""

    resource_group: str = Field(description="Resource group name")
    domain_name: str = Field(description="Domain name")


class RGUserUsageBucketSearchPathParam(BaseRequestModel):
    """Path parameters for RG-scoped user usage bucket search."""

    resource_group: str = Field(description="Resource group name")
    domain_name: str = Field(description="Domain name")
    project_id: UUID = Field(description="Project ID")


# Get Requests (deprecated, use PathParam models above)


class GetDomainFairShareRequest(BaseRequestModel):
    """Request for getting a single domain fair share."""

    resource_group: str = Field(description="Scaling group name")
    domain_name: str = Field(description="Domain name")


class GetProjectFairShareRequest(BaseRequestModel):
    """Request for getting a single project fair share."""

    resource_group: str = Field(description="Scaling group name")
    project_id: UUID = Field(description="Project ID")


class GetUserFairShareRequest(BaseRequestModel):
    """Request for getting a single user fair share."""

    resource_group: str = Field(description="Scaling group name")
    project_id: UUID = Field(description="Project ID")
    user_uuid: UUID = Field(description="User UUID")


# Search Requests


class SearchDomainFairSharesRequest(BaseRequestModel):
    """Request body for searching domain fair shares."""

    filter: DomainFairShareFilter | None = Field(default=None, description="Filter conditions")
    order: list[DomainFairShareOrder] | None = Field(
        default=None, description="Order specifications"
    )
    limit: int = Field(default=50, ge=1, le=1000, description="Maximum items to return")
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


class SearchProjectFairSharesRequest(BaseRequestModel):
    """Request body for searching project fair shares."""

    filter: ProjectFairShareFilter | None = Field(default=None, description="Filter conditions")
    order: list[ProjectFairShareOrder] | None = Field(
        default=None, description="Order specifications"
    )
    limit: int = Field(default=50, ge=1, le=1000, description="Maximum items to return")
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


class SearchUserFairSharesRequest(BaseRequestModel):
    """Request body for searching user fair shares."""

    filter: UserFairShareFilter | None = Field(default=None, description="Filter conditions")
    order: list[UserFairShareOrder] | None = Field(default=None, description="Order specifications")
    limit: int = Field(default=50, ge=1, le=1000, description="Maximum items to return")
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


class SearchDomainUsageBucketsRequest(BaseRequestModel):
    """Request body for searching domain usage buckets."""

    filter: DomainUsageBucketFilter | None = Field(default=None, description="Filter conditions")
    order: list[DomainUsageBucketOrder] | None = Field(
        default=None, description="Order specifications"
    )
    limit: int = Field(default=50, ge=1, le=1000, description="Maximum items to return")
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


class SearchProjectUsageBucketsRequest(BaseRequestModel):
    """Request body for searching project usage buckets."""

    filter: ProjectUsageBucketFilter | None = Field(default=None, description="Filter conditions")
    order: list[ProjectUsageBucketOrder] | None = Field(
        default=None, description="Order specifications"
    )
    limit: int = Field(default=50, ge=1, le=1000, description="Maximum items to return")
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


class SearchUserUsageBucketsRequest(BaseRequestModel):
    """Request body for searching user usage buckets."""

    filter: UserUsageBucketFilter | None = Field(default=None, description="Filter conditions")
    order: list[UserUsageBucketOrder] | None = Field(
        default=None, description="Order specifications"
    )
    limit: int = Field(default=50, ge=1, le=1000, description="Maximum items to return")
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


# Upsert Weight Path Parameters


class UpsertDomainFairShareWeightPathParam(BaseRequestModel):
    """Path parameters for upserting domain fair share weight."""

    resource_group: str = Field(description="Scaling group name")
    domain_name: str = Field(description="Domain name")


class UpsertProjectFairShareWeightPathParam(BaseRequestModel):
    """Path parameters for upserting project fair share weight."""

    resource_group: str = Field(description="Scaling group name")
    project_id: UUID = Field(description="Project ID")


class UpsertUserFairShareWeightPathParam(BaseRequestModel):
    """Path parameters for upserting user fair share weight."""

    resource_group: str = Field(description="Scaling group name")
    project_id: UUID = Field(description="Project ID")
    user_uuid: UUID = Field(description="User UUID")


class UpdateResourceGroupFairShareSpecPathParam(BaseRequestModel):
    """Path parameters for updating resource group fair share spec."""

    resource_group: str = Field(description="Scaling group name")


class GetResourceGroupFairShareSpecPathParam(BaseRequestModel):
    """Path parameters for getting resource group fair share spec."""

    resource_group: str = Field(description="Scaling group name")


# Upsert Weight Request Bodies


class UpsertDomainFairShareWeightRequest(BaseRequestModel):
    """Request body for upserting domain fair share weight.

    Set weight to null to use the resource group's default_weight.
    """

    weight: Decimal | None = Field(
        default=None,
        description=(
            "Priority weight multiplier. Higher weight = higher priority. "
            "Set to null to use resource group's default_weight."
        ),
    )


class UpsertProjectFairShareWeightRequest(BaseRequestModel):
    """Request body for upserting project fair share weight.

    Set weight to null to use the resource group's default_weight.
    """

    domain_name: str = Field(description="Domain name the project belongs to")
    weight: Decimal | None = Field(
        default=None,
        description=(
            "Priority weight multiplier. Higher weight = higher priority. "
            "Set to null to use resource group's default_weight."
        ),
    )


class UpsertUserFairShareWeightRequest(BaseRequestModel):
    """Request body for upserting user fair share weight.

    Set weight to null to use the resource group's default_weight.
    """

    domain_name: str = Field(description="Domain name the user belongs to")
    weight: Decimal | None = Field(
        default=None,
        description=(
            "Priority weight multiplier. Higher weight = higher priority. "
            "Set to null to use resource group's default_weight."
        ),
    )


# Update Resource Group Fair Share Spec


class ResourceWeightEntryInput(BaseModel):
    """A single resource weight entry for fair share calculation.

    Set weight to null to remove the resource type weight (revert to default).
    """

    resource_type: str = Field(description="Resource type identifier (e.g., cpu, mem, cuda.shares)")
    weight: Decimal | None = Field(
        default=None,
        description=(
            "Weight multiplier for this resource type. "
            "Set to null to remove (revert to default weight)."
        ),
    )


class UpdateResourceGroupFairShareSpecRequest(BaseRequestModel):
    """Request body for updating resource group fair share spec (partial update).

    All fields are optional. Only provided fields are updated; others retain existing values.
    """

    half_life_days: int | None = Field(
        default=None,
        description="Half-life for exponential decay in days. Leave null to keep existing value.",
    )
    lookback_days: int | None = Field(
        default=None,
        description="Total lookback period in days. Leave null to keep existing value.",
    )
    decay_unit_days: int | None = Field(
        default=None,
        description="Granularity of decay buckets in days. Leave null to keep existing value.",
    )
    default_weight: Decimal | None = Field(
        default=None,
        description="Default weight for entities. Leave null to keep existing value.",
    )
    resource_weights: list[ResourceWeightEntryInput] | None = Field(
        default=None,
        description=(
            "Resource weights for fair share calculation. "
            "Each entry specifies a resource type and its weight multiplier. "
            "Only provided resource types are updated (partial update). "
            "Set weight to null to remove that resource type (revert to default). "
            "Leave the entire list null to keep all existing values."
        ),
    )


# Bulk Upsert Weight Input Types


class DomainWeightEntryInput(BaseRequestModel):
    """Single domain weight entry for bulk upsert."""

    domain_name: str = Field(description="Domain name")
    weight: Decimal | None = Field(description="Fair share weight (null for default)")


class BulkUpsertDomainFairShareWeightRequest(BaseRequestModel):
    """Request to bulk upsert domain fair share weights."""

    resource_group: str = Field(description="Scaling group name")
    inputs: list[DomainWeightEntryInput] = Field(description="List of domain weights to upsert")


class ProjectWeightEntryInput(BaseRequestModel):
    """Single project weight entry for bulk upsert."""

    project_id: UUID = Field(description="Project ID")
    domain_name: str = Field(description="Domain name for context")
    weight: Decimal | None = Field(description="Fair share weight (null for default)")


class BulkUpsertProjectFairShareWeightRequest(BaseRequestModel):
    """Request to bulk upsert project fair share weights."""

    resource_group: str = Field(description="Scaling group name")
    inputs: list[ProjectWeightEntryInput] = Field(description="List of project weights to upsert")


class UserWeightEntryInput(BaseRequestModel):
    """Single user weight entry for bulk upsert."""

    user_uuid: UUID = Field(description="User UUID")
    project_id: UUID = Field(description="Project ID")
    domain_name: str = Field(description="Domain name for context")
    weight: Decimal | None = Field(description="Fair share weight (null for default)")


class BulkUpsertUserFairShareWeightRequest(BaseRequestModel):
    """Request to bulk upsert user fair share weights."""

    resource_group: str = Field(description="Scaling group name")
    inputs: list[UserWeightEntryInput] = Field(description="List of user weights to upsert")



from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any, cast, override

from ai.backend.common.bgtask.types import BgtaskNameBase
from ai.backend.common.exception import (
    BgtaskNotRegisteredError,
)

from .base import (
    BaseBackgroundTaskHandler,
    BaseBackgroundTaskManifest,
    BaseBackgroundTaskResult,
)


class _TaskExecutor(ABC):
    @abstractmethod
    async def revive_task(
        self, manifest_dict: Mapping[str, Any]
    ) -> BaseBackgroundTaskResult | None:
        """
        Revive the background task with the provided manifest dictionary.
        This method should be implemented by subclasses to provide
        the specific revival logic.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    async def execute_new_task(
        self, manifest: BaseBackgroundTaskManifest
    ) -> BaseBackgroundTaskResult | None:
        """
        Execute the background task with the provided manifest.
        This method should be implemented by subclasses to provide
        the specific execution logic.
        """
        raise NotImplementedError("Subclasses must implement this method")


class _TaskDefinition[
    TManifest: BaseBackgroundTaskManifest,
    TResult: BaseBackgroundTaskResult | None,
](_TaskExecutor):
    _handler: BaseBackgroundTaskHandler[TManifest, TResult]
    _task_name: BgtaskNameBase

    def __init__(self, handler: BaseBackgroundTaskHandler[TManifest, TResult]) -> None:
        self._handler = handler
        self._task_name = handler.name()

    def task_name(self) -> BgtaskNameBase:
        return self._task_name

    @override
    async def revive_task(self, manifest_dict: Mapping[str, Any]) -> TResult:
        manifest_instance = self._handler.manifest_type().model_validate(manifest_dict)
        return await self._handler.execute(manifest_instance)

    @override
    async def execute_new_task(self, manifest: BaseBackgroundTaskManifest) -> TResult:
        if not isinstance(manifest, self._handler.manifest_type()):
            raise TypeError(
                f"Expected manifest of type {self._handler.manifest_type()}, got {type(manifest)}"
            )
        return await self._handler.execute(manifest)


class BackgroundTaskHandlerRegistry:
    _executor_registry: dict[str, _TaskDefinition]  # type: ignore[type-arg]

    def __init__(self) -> None:
        self._executor_registry = {}

    def register(self, handler: BaseBackgroundTaskHandler) -> None:  # type: ignore[type-arg]
        self._executor_registry[handler.name().value] = _TaskDefinition(
            handler=handler,
        )

    def get_task_name(self, name: str) -> BgtaskNameBase:
        """Get BgtaskNameBase instance from string name."""
        try:
            definition = self._executor_registry[name]
        except KeyError as e:
            raise BgtaskNotRegisteredError(f"Task '{name}' is not registered.") from e
        return definition.task_name()

    async def revive_task(
        self, name: str, manifest_dict: Mapping[str, Any]
    ) -> BaseBackgroundTaskResult | None:
        try:
            definition = self._executor_registry[name]
        except KeyError as e:
            raise BgtaskNotRegisteredError(f"Task '{name}' is not registered.") from e
        return cast(BaseBackgroundTaskResult | None, await definition.revive_task(manifest_dict))

    async def execute_new_task(
        self, name: BgtaskNameBase, manifest: BaseBackgroundTaskManifest
    ) -> BaseBackgroundTaskResult | None:
        try:
            definition = self._executor_registry[name.value]
        except KeyError as e:
            raise BgtaskNotRegisteredError(f"Task '{name}' is not registered.") from e
        return cast(BaseBackgroundTaskResult | None, await definition.execute_new_task(manifest))



"""JWT token types and claims for GraphQL Federation authentication."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Any

from ai.backend.common.types import AccessKey


@dataclass(frozen=True)
class JWTUserContext:
    """
    User context data for JWT token generation.

    This dataclass encapsulates minimal user information needed to generate
    a JWT token. Additional user information (user_id, domain_name, is_admin, is_superadmin)
    should be retrieved from the user table during authentication.

    Attributes:
        access_key: User's access key
        role: User's role ("admin", "user", or "superadmin")
    """

    access_key: AccessKey
    role: str


@dataclass(frozen=True)
class JWTClaims:
    """
    JWT token payload for GraphQL Federation authentication.

    This dataclass represents the claims contained in a JWT token used for
    authenticating GraphQL requests through Hive Router.

    Contains minimal user information. Additional user information (user_id, domain_name,
    is_admin, is_superadmin) should be retrieved from the user table during authentication.

    Attributes:
        exp: Expiration time (UTC)
        iat: Issued at time (UTC)
        access_key: User's access key
        role: User role ("admin", "user", or "superadmin")
    """

    # Standard JWT claims (RFC 7519)
    exp: datetime
    iat: datetime

    # Backend.AI specific claims
    access_key: AccessKey
    role: str

    def to_dict(self) -> dict[str, Any]:
        """
        Convert JWTClaims to a dictionary suitable for JWT payload.

        Datetime objects are converted to Unix timestamps (integers) as required
        by the JWT standard.

        Returns:
            Dictionary representation of claims with timestamps as integers.
        """
        return {
            "exp": int(self.exp.timestamp()),
            "iat": int(self.iat.timestamp()),
            "access_key": str(self.access_key),
            "role": self.role,
        }

    @classmethod
    def from_dict(cls, payload: dict[str, Any]) -> JWTClaims:
        """
        Parse JWT payload dictionary to JWTClaims.

        Converts Unix timestamps back to datetime objects and validates
        the structure of the payload.

        Args:
            payload: Dictionary containing JWT claims

        Returns:
            JWTClaims instance

        Raises:
            KeyError: If required claims are missing
            ValueError: If claim values are invalid
        """
        return cls(
            exp=datetime.fromtimestamp(payload["exp"], tz=UTC),
            iat=datetime.fromtimestamp(payload["iat"], tz=UTC),
            access_key=AccessKey(payload["access_key"]),
            role=payload["role"],
        )



from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class HealthCheckError(BackendAIError):
    """
    Base exception for all health check failures.

    Subclasses should inherit this along with appropriate web.HTTPxxx classes
    and implement the error_code() method.
    """

    pass


class HttpHealthCheckError(HealthCheckError, web.HTTPServiceUnavailable):
    """Raised when HTTP endpoint health check fails."""

    error_type = "https://api.backend.ai/probs/http-health-check-failed"
    error_title = "HTTP health check failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.HEALTH_CHECK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class ValkeyHealthCheckError(HealthCheckError, web.HTTPServiceUnavailable):
    """Raised when Valkey/Redis health check fails."""

    error_type = "https://api.backend.ai/probs/valkey-health-check-failed"
    error_title = "Valkey health check failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.HEALTH_CHECK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class EtcdHealthCheckError(HealthCheckError, web.HTTPServiceUnavailable):
    """Raised when etcd health check fails."""

    error_type = "https://api.backend.ai/probs/etcd-health-check-failed"
    error_title = "Etcd health check failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.HEALTH_CHECK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class DatabaseHealthCheckError(HealthCheckError, web.HTTPServiceUnavailable):
    """Raised when database health check fails."""

    error_type = "https://api.backend.ai/probs/database-health-check-failed"
    error_title = "Database health check failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.HEALTH_CHECK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class DockerHealthCheckError(HealthCheckError, web.HTTPServiceUnavailable):
    """Raised when Docker health check fails."""

    error_type = "https://api.backend.ai/probs/docker-health-check-failed"
    error_title = "Docker health check failed"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.HEALTH_CHECK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class HealthCheckerAlreadyRegistered(BackendAIError, web.HTTPConflict):
    """
    Raised when attempting to register a health checker that is already registered.
    """

    error_type = "https://api.backend.ai/probs/health-checker-already-registered"
    error_title = "Health checker already registered."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.HEALTH_CHECK,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class HealthCheckerNotFound(BackendAIError, web.HTTPNotFound):
    """
    Raised when attempting to access a health checker that is not registered.
    """

    error_type = "https://api.backend.ai/probs/health-checker-not-found"
    error_title = "Health checker not found."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.HEALTH_CHECK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )



"""
Request DTOs for RBAC system.
Shared between Client SDK and Manager API.
"""

from __future__ import annotations

from uuid import UUID

from pydantic import Field

from ai.backend.common.api_handlers import SENTINEL, BaseRequestModel, Sentinel
from ai.backend.common.data.permission.types import ScopeType
from ai.backend.common.dto.manager.defs import DEFAULT_PAGE_LIMIT, MAX_PAGE_LIMIT
from ai.backend.common.dto.manager.query import StringFilter

from .types import (
    AssignedUserOrderField,
    EntityType,
    OperationType,
    OrderDirection,
    PermissionStatus,
    RoleOrderField,
    RoleSource,
    RoleStatus,
    ScopeOrderField,
)

__all__ = (
    "AssignRoleRequest",
    "AssignedUserFilter",
    "AssignedUserOrder",
    "CreateObjectPermissionRequest",
    "CreatePermissionRequest",
    "CreateRoleRequest",
    "RevokeRoleRequest",
    "RoleFilter",
    "RoleOrder",
    "ScopeFilter",
    "ScopeOrder",
    "SearchEntitiesRequest",
    "SearchRolesRequest",
    "SearchScopesRequest",
    "SearchUsersAssignedToRoleRequest",
    "StringFilter",
    "UpdateRoleRequest",
)


class CreateRoleRequest(BaseRequestModel):
    """Request to create a role."""

    name: str = Field(description="Role name")
    source: RoleSource = Field(default=RoleSource.CUSTOM, description="Role source")
    status: RoleStatus = Field(default=RoleStatus.ACTIVE, description="Role status")
    description: str | None = Field(default=None, description="Role description")


class UpdateRoleRequest(BaseRequestModel):
    """Request to update a role."""

    name: str | None = Field(default=None, description="Updated role name")
    source: RoleSource | None = Field(default=None, description="Updated role source")
    status: RoleStatus | None = Field(default=None, description="Updated role status")
    description: str | Sentinel | None = Field(
        default=SENTINEL, description="Updated role description"
    )


class DeleteRoleRequest(BaseRequestModel):
    """Request to delete a role."""

    role_id: UUID = Field(description="Role ID to delete")


class PurgeRoleRequest(BaseRequestModel):
    """Request to purge a role."""

    role_id: UUID = Field(description="Role ID to purge")


class AssignRoleRequest(BaseRequestModel):
    """Request to assign a role to a user."""

    user_id: UUID = Field(description="User ID to assign the role to")
    role_id: UUID = Field(description="Role ID to assign")
    granted_by: UUID | None = Field(
        default=None, description="User ID who granted this role assignment"
    )


class RevokeRoleRequest(BaseRequestModel):
    """Request to revoke a role from a user."""

    user_id: UUID = Field(description="User ID to revoke the role from")
    role_id: UUID = Field(description="Role ID to revoke")


class RoleFilter(BaseRequestModel):
    """Filter for roles."""

    name: StringFilter | None = Field(default=None, description="Filter by name")
    sources: list[RoleSource] | None = Field(default=None, description="Filter by role sources")
    statuses: list[RoleStatus] | None = Field(default=None, description="Filter by role statuses")


class RoleOrder(BaseRequestModel):
    """Order specification for roles."""

    field: RoleOrderField = Field(description="Field to order by")
    direction: OrderDirection = Field(default=OrderDirection.ASC, description="Order direction")


class SearchRolesRequest(BaseRequestModel):
    """Request body for searching roles with filters, orders, and pagination."""

    filter: RoleFilter | None = Field(default=None, description="Filter conditions")
    order: list[RoleOrder] | None = Field(default=None, description="Order specifications")
    limit: int = Field(
        default=DEFAULT_PAGE_LIMIT, ge=1, le=MAX_PAGE_LIMIT, description="Maximum items to return"
    )
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


class AssignedUserFilter(BaseRequestModel):
    """Filter for assigned users."""

    username: StringFilter | None = Field(default=None, description="Filter by username")
    email: StringFilter | None = Field(default=None, description="Filter by email")
    granted_by: UUID | None = Field(default=None, description="Filter by 'granted_by' user ID")


class AssignedUserOrder(BaseRequestModel):
    """Order specification for assigned users."""

    field: AssignedUserOrderField = Field(description="Field to order by")
    direction: OrderDirection = Field(default=OrderDirection.ASC, description="Order direction")


class SearchUsersAssignedToRoleRequest(BaseRequestModel):
    """Request body for searching users assigned to a specific role."""

    filter: AssignedUserFilter | None = Field(default=None, description="Filter conditions")
    order: list[AssignedUserOrder] | None = Field(default=None, description="Order specifications")
    limit: int = Field(
        default=DEFAULT_PAGE_LIMIT, ge=1, le=MAX_PAGE_LIMIT, description="Maximum items to return"
    )
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


class CreatePermissionRequest(BaseRequestModel):
    """Request to create a permission."""

    role_id: UUID = Field(description="Role ID for the permission")
    scope_type: ScopeType = Field(description="Scope type for the permission")
    scope_id: str = Field(description="Scope ID for the permission")
    entity_type: EntityType = Field(description="Entity type for the permission")
    operation: OperationType = Field(description="Operation type for the permission")


class CreateObjectPermissionRequest(BaseRequestModel):
    """Request to create an object permission for a role."""

    role_id: UUID = Field(description="Role ID to add the object permission to")
    entity_type: EntityType = Field(description="Entity type for the object permission")
    entity_id: str = Field(description="Entity ID (e.g., project_id, user_id)")
    operation: OperationType = Field(description="Operation type for the object permission")
    status: PermissionStatus = Field(
        default=PermissionStatus.ACTIVE, description="Permission status"
    )


class ScopeFilter(BaseRequestModel):
    """Filter for scopes."""

    name: StringFilter | None = Field(default=None, description="Filter by name")


class ScopeOrder(BaseRequestModel):
    """Order specification for scopes."""

    field: ScopeOrderField = Field(description="Field to order by")
    direction: OrderDirection = Field(default=OrderDirection.ASC, description="Order direction")


class SearchScopesRequest(BaseRequestModel):
    """Request body for searching scopes with filters and pagination."""

    filter: ScopeFilter | None = Field(default=None, description="Filter conditions")
    order: list[ScopeOrder] | None = Field(default=None, description="Order specifications")
    limit: int = Field(
        default=DEFAULT_PAGE_LIMIT, ge=1, le=MAX_PAGE_LIMIT, description="Maximum items to return"
    )
    offset: int = Field(default=0, ge=0, description="Number of items to skip")


class SearchEntitiesRequest(BaseRequestModel):
    """Request body for searching entities within a scope."""

    limit: int = Field(
        default=DEFAULT_PAGE_LIMIT, ge=1, le=MAX_PAGE_LIMIT, description="Maximum items to return"
    )
    offset: int = Field(default=0, ge=0, description="Number of items to skip")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Self
from uuid import UUID

from ai.backend.common.types import CIStrEnum


class UserRole(CIStrEnum):
    """
    User's role.
    """

    SUPERADMIN = "superadmin"
    ADMIN = "admin"
    USER = "user"
    MONITOR = "monitor"


@dataclass(frozen=True)
class UserData:
    user_id: UUID
    is_authorized: bool
    is_admin: bool
    is_superadmin: bool
    role: UserRole
    domain_name: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        """
        Create UserData from a dictionary.

        Handles conversion of string role to UserRole enum.
        """
        if "role" in data and isinstance(data["role"], str):
            data = {**data, "role": UserRole(data["role"])}
        return cls(**data)



from __future__ import annotations

import asyncio
from collections.abc import Sequence
from datetime import UTC, datetime
from http import HTTPMethod

import aiohttp

from ai.backend.common.health_checker.abc import StaticServiceHealthChecker
from ai.backend.common.health_checker.types import (
    API,
    ComponentHealthStatus,
    ComponentId,
    ServiceGroup,
    ServiceHealth,
)


class HttpHealthChecker(StaticServiceHealthChecker):
    """
    Health checker for HTTP endpoints.

    Performs HTTP requests to the specified URL and considers the endpoint
    healthy if it returns one of the expected status codes.
    """

    _url: str
    _component_id: ComponentId
    _timeout: float
    _session: aiohttp.ClientSession
    _expected_status_codes: frozenset[int]
    _method: HTTPMethod

    def __init__(
        self,
        url: str,
        component_id: ComponentId,
        session: aiohttp.ClientSession,
        timeout: float = 5.0,
        expected_status_codes: Sequence[int] = (200,),
        method: HTTPMethod = HTTPMethod.GET,
    ) -> None:
        """
        Initialize HttpHealthChecker.

        Args:
            url: The HTTP endpoint URL to check
            component_id: Component identifier for this endpoint
            session: aiohttp ClientSession to use for requests
            timeout: Timeout in seconds for the health check request
            expected_status_codes: HTTP status codes considered healthy (default: [200])
            method: HTTP method to use for health check (default: GET)
        """
        self._url = url
        self._component_id = component_id
        self._timeout = timeout
        self._session = session
        self._expected_status_codes = frozenset(expected_status_codes)
        self._method = method

    @property
    def target_service_group(self) -> ServiceGroup:
        """The service group this checker monitors."""
        return API

    async def check_service(self) -> ServiceHealth:
        """
        Check HTTP endpoint health by performing an HTTP request.

        Returns:
            ServiceHealth containing status for the HTTP endpoint
        """
        check_time = datetime.now(UTC)

        try:
            async with asyncio.timeout(self._timeout):
                async with self._session.request(self._method.value, self._url) as response:
                    if response.status not in self._expected_status_codes:
                        status = ComponentHealthStatus(
                            is_healthy=False,
                            last_checked_at=check_time,
                            error_message=f"HTTP {self._method} {self._url} returned status {response.status}, expected one of {sorted(self._expected_status_codes)}",
                        )
                    else:
                        status = ComponentHealthStatus(
                            is_healthy=True,
                            last_checked_at=check_time,
                            error_message=None,
                        )
        except TimeoutError:
            status = ComponentHealthStatus(
                is_healthy=False,
                last_checked_at=check_time,
                error_message=f"HTTP health check timed out after {self._timeout}s: {self._method} {self._url}",
            )
        except aiohttp.ClientError as e:
            status = ComponentHealthStatus(
                is_healthy=False,
                last_checked_at=check_time,
                error_message=f"HTTP health check failed for {self._method} {self._url}: {e}",
            )
        except Exception as e:
            status = ComponentHealthStatus(
                is_healthy=False,
                last_checked_at=check_time,
                error_message=f"Unexpected error during HTTP health check for {self._method} {self._url}: {e}",
            )

        return ServiceHealth(results={self._component_id: status})

    @property
    def timeout(self) -> float:
        """The timeout for the health check in seconds."""
        return self._timeout



"""Common permission types shared between manager and client."""

from __future__ import annotations

import enum


class PermissionStatus(enum.StrEnum):
    ACTIVE = "active"
    # 'inactive' status is used when the permission is temporarily disabled
    INACTIVE = "inactive"
    # 'deleted' status is used when the permission is permanently removed
    DELETED = "deleted"


class RoleStatus(enum.StrEnum):
    """Status of a role."""

    ACTIVE = "active"
    # 'inactive' status is used when the role is temporarily disabled
    INACTIVE = "inactive"
    # 'deleted' status is used when the role is permanently removed
    DELETED = "deleted"


class RoleSource(enum.StrEnum):
    """Definition source of the role."""

    SYSTEM = "system"  # System-defined role, e.g., default roles
    CUSTOM = "custom"  # Custom role defined


class OperationType(enum.StrEnum):
    CREATE = "create"
    READ = "read"
    UPDATE = "update"
    SOFT_DELETE = "soft-delete"
    HARD_DELETE = "hard-delete"
    GRANT_ALL = "grant:all"  # Allow user to grant all permissions, including grant of grant
    GRANT_READ = "grant:read"
    GRANT_UPDATE = "grant:update"
    GRANT_SOFT_DELETE = "grant:soft-delete"
    GRANT_HARD_DELETE = "grant:hard-delete"

    @classmethod
    def owner_operations(cls) -> set[OperationType]:
        """
        Returns a set of operations that are considered owner operations.
        Owner operations are those that allow full control over an entity.
        """
        return {op for op in cls}

    @classmethod
    def admin_operations(cls) -> set[OperationType]:
        """
        Returns a set of operations that are considered admin operations.
        Admin operations are those that allow management of entities, including creation and deletion.
        """
        return {op for op in cls}

    @classmethod
    def member_operations(cls) -> set[OperationType]:
        """
        Returns a set of operations that are considered member operations.
        Member operations are those that allow read access.
        """
        return {
            cls.READ,
        }


class EntityType(enum.StrEnum):
    # === RBAC scope/resource types (original 12) ===
    USER = "user"
    PROJECT = "project"
    DOMAIN = "domain"

    VFOLDER = "vfolder"
    IMAGE = "image"
    SESSION = "session"

    ARTIFACT = "artifact"
    ARTIFACT_REGISTRY = "artifact_registry"
    APP_CONFIG = "app_config"
    NOTIFICATION_CHANNEL = "notification_channel"
    NOTIFICATION_RULE = "notification_rule"
    MODEL_DEPLOYMENT = "model_deployment"

    # === Standalone entity types ===
    AGENT = "agent"
    AUTH = "auth"
    AUDIT_LOG = "audit_log"
    CONTAINER_METRIC = "container_metric"
    CONTAINER_METRIC_METADATA = "container_metric_metadata"
    CONTAINER_REGISTRY = "container_registry"
    DEPLOYMENT = "deployment"
    ERROR_LOG = "error_log"
    EXPORT = "export"
    GROUP = "group"
    MODEL_SERVICE = "model_service"
    NOTIFICATION = "notification"
    OBJECT_PERMISSION = "object_permission"
    OBJECT_STORAGE = "object_storage"
    PERMISSION = "permission"
    RESOURCE_GROUP = "resource_group"
    RESOURCE_PRESET = "resource_preset"
    ROLE = "role"
    STORAGE_NAMESPACE = "storage_namespace"
    VFS_STORAGE = "vfs_storage"

    # === Parent:child sub-entity types (`:` separator) ===
    # Domain/Project/User sub-resources
    DOMAIN_FAIR_SHARE = "domain:fair_share"
    PROJECT_FAIR_SHARE = "project:fair_share"
    USER_FAIR_SHARE = "user:fair_share"
    DOMAIN_USAGE_BUCKET = "domain:usage_bucket"
    PROJECT_USAGE_BUCKET = "project:usage_bucket"
    USER_USAGE_BUCKET = "user:usage_bucket"
    KEYPAIR_RESOURCE_POLICY = "keypair:resource_policy"
    PROJECT_RESOURCE_POLICY = "project:resource_policy"
    USER_RESOURCE_POLICY = "user:resource_policy"
    # App config sub
    APP_CONFIG_DOMAIN = "app_config:domain"
    APP_CONFIG_USER = "app_config:user"
    # Session sub
    SESSION_KERNEL = "session:kernel"
    SESSION_FILE = "session:file"
    SESSION_DIRECTORY = "session:directory"
    SESSION_APP_SERVICE = "session:app_service"
    SESSION_COMMIT = "session:commit"
    SESSION_STATUS_HISTORY = "session:status_history"
    SESSION_ABUSING_REPORT = "session:abusing_report"
    SESSION_CONTAINER_LOG = "session:container_log"
    SESSION_DEPENDENCY_GRAPH = "session:dependency_graph"
    SESSION_DIRECT_ACCESS = "session:direct_access"
    SESSION_HISTORY = "session:history"
    SESSION_SCOPED_HISTORY = "session:scoped_history"
    # Deployment sub
    DEPLOYMENT_REPLICA = "deployment:replica"
    DEPLOYMENT_ROUTE = "deployment:route"
    DEPLOYMENT_ACCESS_TOKEN = "deployment:access_token"
    DEPLOYMENT_AUTO_SCALING_RULE = "deployment:auto_scaling_rule"
    DEPLOYMENT_MODEL_REVISION = "deployment:model_revision"
    DEPLOYMENT_REVISION = "deployment:revision"
    DEPLOYMENT_POLICY = "deployment:policy"
    DEPLOYMENT_HISTORY = "deployment:history"
    DEPLOYMENT_SCOPED_HISTORY = "deployment:scoped_history"
    DEPLOYMENT_ERROR = "deployment:error"
    DEPLOYMENT_TOKEN = "deployment:token"
    # Image sub
    IMAGE_ALIAS = "image:alias"
    IMAGE_TAG = "image:tag"
    IMAGE_PRELOAD = "image:preload"
    IMAGE_SCAN = "image:scan"
    IMAGE_AGENT = "image:agent"
    IMAGE_RESOURCE_LIMIT = "image:resource_limit"
    # Artifact sub
    ARTIFACT_REVISION = "artifact:revision"
    ARTIFACT_SCAN = "artifact:scan"
    ARTIFACT_MODEL = "artifact:model"
    ARTIFACT_IMPORT = "artifact:import"
    ARTIFACT_DOWNLOAD = "artifact:download"
    ARTIFACT_README = "artifact:readme"
    ARTIFACT_VERIFICATION = "artifact:verification"
    ARTIFACT_REVISION_STORAGE_LINK = "artifact:revision:storage_link"
    # VFolder sub
    VFOLDER_FILE = "vfolder:file"
    VFOLDER_DIRECTORY = "vfolder:directory"
    VFOLDER_INVITATION = "vfolder:invitation"
    # Resource group sub
    RESOURCE_GROUP_DOMAIN = "resource_group:domain"
    RESOURCE_GROUP_KEYPAIR = "resource_group:keypair"
    RESOURCE_GROUP_USER_GROUP = "resource_group:user_group"
    RESOURCE_GROUP_FAIR_SHARE = "resource_group:fair_share"
    RESOURCE_GROUP_RESOURCE = "resource_group:resource"
    # Role sub
    ROLE_ASSIGNMENT = "role:assignment"
    ROLE_USER = "role:user"
    ROLE_ENTITY = "role:entity"
    ROLE_SCOPE = "role:scope"
    ROLE_PERMISSION = "role:permission"
    ROLE_ENTITY_TYPE = "role:entity_type"
    ROLE_SCOPE_TYPE = "role:scope_type"
    # Container registry sub
    CONTAINER_REGISTRY_IMAGE = "container_registry:image"
    # Route sub
    ROUTE_HISTORY = "route:history"
    ROUTE_SCOPED_HISTORY = "route:scoped_history"
    # Auth sub
    AUTH_TOKEN = "auth:token"
    AUTH_ACCOUNT = "auth:account"
    AUTH_SESSION = "auth:session"
    AUTH_PASSWORD = "auth:password"
    AUTH_PROFILE = "auth:profile"
    AUTH_SSH_KEYPAIR = "auth:ssh_keypair"
    # Notification sub
    NOTIFICATION_EVENT = "notification:event"
    NOTIFICATION_CHANNEL_VALIDATION = "notification:channel_validation"
    NOTIFICATION_RULE_VALIDATION = "notification:rule_validation"
    # Group sub
    GROUP_USAGE = "group:usage"
    # User sub
    USER_STATS = "user:stats"
    # Agent sub
    AGENT_WATCHER = "agent:watcher"
    AGENT_REGISTRY = "agent:registry"
    # Resource preset sub
    RESOURCE_PRESET_CHECK = "resource_preset:check"
    # Permission sub
    PERMISSION_CHECK = "permission:check"
    # Export sub
    EXPORT_REPORT = "export:report"

    @classmethod
    def _scope_types(cls) -> set[EntityType]:
        """
        Returns a set of entity types that are considered scope types.
        """
        return {cls.USER, cls.PROJECT, cls.DOMAIN}

    @classmethod
    def _resource_types(cls) -> set[EntityType]:
        """
        Returns a set of entity types that are considered resource types.
        """
        return {
            cls.VFOLDER,
            cls.IMAGE,
            cls.SESSION,
            cls.ARTIFACT,
            cls.ARTIFACT_REGISTRY,
            cls.APP_CONFIG,
            cls.NOTIFICATION_CHANNEL,
            cls.NOTIFICATION_RULE,
            cls.MODEL_DEPLOYMENT,
        }

    @classmethod
    def owner_accessible_entity_types_in_user(cls) -> set[EntityType]:
        """
        Returns a set of entity types that are accessible by owner roles in user scope.
        """
        return cls._resource_types()

    @classmethod
    def admin_accessible_entity_types_in_project(cls) -> set[EntityType]:
        """
        Returns a set of entity types that are accessible by admin roles.
        """
        return {*cls._resource_types(), cls.USER}

    @classmethod
    def admin_accessible_entity_types_in_domain(cls) -> set[EntityType]:
        """
        Returns a set of entity types that are accessible by admin roles.
        """
        return {*cls._resource_types(), cls.USER}

    @classmethod
    def member_accessible_entity_types_in_project(cls) -> set[EntityType]:
        """
        Returns a set of entity types that are accessible by member roles.
        """
        return {*cls._resource_types(), cls.USER}

    @classmethod
    def member_accessible_entity_types_in_domain(cls) -> set[EntityType]:
        """
        Returns a set of entity types that are accessible by member roles.
        """
        return {*cls._resource_types(), cls.USER}


class FieldType(enum.StrEnum):
    """Field types for RBAC field-scoped entities.

    Fields are sub-resources that belong to a parent entity.
    Unlike EntityType which represents scope-scoped entities,
    FieldType represents entity-scoped sub-entities.
    """

    KERNEL = "kernel"
    MODEL_REVISION = "model_revision"


class ScopeType(enum.StrEnum):
    # === Organization/permission scopes (original) ===
    DOMAIN = "domain"
    PROJECT = "project"
    USER = "user"
    GLOBAL = "global"

    RESOURCE_GROUP = "resource_group"
    CONTAINER_REGISTRY = "container_registry"
    ARTIFACT_REGISTRY = "artifact_registry"
    STORAGE_HOST = "storage_host"

    # === Entity-level scopes ===
    SESSION = "session"
    DEPLOYMENT = "deployment"
    VFOLDER = "vfolder"
    IMAGE = "image"
    ARTIFACT = "artifact"
    ARTIFACT_REVISION = "artifact_revision"
    ROLE = "role"


GLOBAL_SCOPE_ID = "global"


class RelationType(enum.StrEnum):
    """Classification of parent-child entity edges in BEP-1048.

    AUTO: Composition edge with permission delegation from parent.
    REF: Read-only reference edge with no permission delegation.
    """

    AUTO = "auto"
    REF = "ref"



from .etcd import EtcdConfig
from .otel import OTELConfig
from .pyroscope import PyroscopeConfig
from .service_discovery import ServiceDiscoveryConfig, ServiceEndpointConfig

__all__ = (
    "EtcdConfig",
    "OTELConfig",
    "PyroscopeConfig",
    "ServiceDiscoveryConfig",
    "ServiceEndpointConfig",
)
